# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


import datetime
from typing import Hashable, Any, Optional, Dict, List, Mapping, Union
import itertools

import numpy as np
import pandas as pd
import pyproj
import xarray as xr

import xcube
from xcube.core.gridmapping import CRS_CRS84, GridMapping
from xcube.server.api import ApiError
from xcube.server.api import ServerConfig
from xcube.util.jsonencoder import to_json_value
from xcube.util.jsonschema import JsonObjectSchema, JsonSchema
from .config import DEFAULT_CATALOG_DESCRIPTION, DEFAULT_FEATURE_ID
from .config import DEFAULT_CATALOG_ID
from .config import DEFAULT_CATALOG_TITLE
from .config import DEFAULT_COLLECTION_DESCRIPTION
from .config import DEFAULT_COLLECTION_ID
from .config import DEFAULT_COLLECTION_TITLE
from .config import PATH_PREFIX
from .context import StacContext
from ..coverages.controllers import get_crs_from_dataset
from ...datasets.context import DatasetsContext

_REL_DOMAINSET = 'http://www.opengis.net/def/rel/ogc/1.0/coverage-domainset'
_REL_RANGETYPE = 'http://www.opengis.net/def/rel/ogc/1.0/coverage-rangetype'
_REL_SCHEMA = 'http://www.opengis.net/def/rel/ogc/1.0/schema'
_JSON_SCHEMA_METASCHEMA = 'https://json-schema.org/draft/2020-12/schema'

STAC_VERSION = '1.0.0'
STAC_EXTENSIONS = [
    "https://stac-extensions.github.io/datacube/v2.1.0/schema.json"
]

# Maximum number of values allowed for the "values" field
# of a value of "datacube:dimensions" or "datacube:variables":
_MAX_NUM_VALUES = 1000

_CONFORMANCE = (
    ['https://api.geodatacube.example/1.0.0-beta']
    + [
        f'https://api.stacspec.org/v1.0.0/{part}'
        for part in ['core', 'collections', 'ogcapi-features']
    ]
    + [
        f'http://www.opengis.net/spec/ogcapi-{ogcapi}/conf/{part}'
        for ogcapi, parts in [
            ('common-1/1.0', ['core', 'landing-page', 'json', 'oas30']),
            ('common-2/0.0', ['collections']),
            ('features-1/1.0', ['core', 'oas30', 'html', 'geojson']),
            (
                'coverages-1/0.0',
                [
                    'core', 'scaling', 'subsetting', 'fieldselection',
                    'crs', 'geotiff', 'netcdf', 'oas30'
                ],
            ),
        ]
        for part in parts
    ]
)


# noinspection PyUnusedLocal
def get_root(ctx: DatasetsContext, base_url: str):
    """Return content for the STAC/OGC root endpoint (a STAC catalogue)

    :param ctx: the datasets context
    :param base_url: the base URL of the server
    :return: content for the root endpoint
    """
    c_id, c_title, c_description = _get_catalog_metadata(ctx.config)

    # If OGC API - Coverages is present, the STAC controller lists the
    # Coverages endpoints along with its own.
    endpoint_lists = [
        a.endpoints()
        for a in ctx.apis
        if a.name in {'ows.stac', 'ows.coverages'}
    ]
    endpoints = list(itertools.chain.from_iterable(endpoint_lists))
    for endpoint in endpoints:
        endpoint['path'] = endpoint['path'][len(PATH_PREFIX):]

    return {
        "type": "Catalog",
        "stac_version": STAC_VERSION,
        "conformsTo": _CONFORMANCE,
        "id": c_id,
        "title": c_title,
        "description": c_description,
        "api_version": "1.0.0",
        "backend_version": xcube.__version__,
        "gdc_version": "1.0.0-beta",
        "endpoints": endpoints,
        "links": [
            _root_link(base_url),
            {
                "rel": "self",
                "href": f'{base_url}{PATH_PREFIX}',
                "type": "application/json",
                "title": "this document"
            },
            {
                "rel": "service-desc",
                "href": f'{base_url}/openapi.json',
                "type": "application/vnd.oai.openapi+json;version=3.0",
                "title": "the API definition"
            },
            {
                "rel": "service-doc",
                "href": f'{base_url}/openapi.html',
                "type": "text/html",
                "title": "the API documentation"
            },
            {
                "rel": "conformance",
                "href": f'{base_url}{PATH_PREFIX}/conformance',
                "type": "application/json",
                "title": "OGC API conformance classes"
                " implemented by this server"
            },
            {
                "rel": "data",
                "href": f'{base_url}{PATH_PREFIX}/collections',
                "type": "application/json",
                "title": "Information about the feature collections"
            },
            {
                "rel": "search",
                "href": f'{base_url}{PATH_PREFIX}/search',
                "type": "application/json",
                "title": "Search across feature collections"
            },
            {
                "rel": "child",
                "href": f'{base_url}{PATH_PREFIX}/collections/datacubes',
                "type": "application/json",
                "title": DEFAULT_COLLECTION_DESCRIPTION
            }
        ],
    }


def _root_link(base_url):
    return {
        "rel": "root",
        "href": f'{base_url}{PATH_PREFIX}',
        "type": "application/json",
        "title": "root of the OGC API and STAC catalog"
    }


# noinspection PyUnusedLocal
def get_conformance() -> dict[str, list[str]]:
    """Return conformance data for this API implementation

    :return: a dictionary containing a list of conformance specifiers
    """
    return {"conformsTo": _CONFORMANCE}


def get_collections(ctx: StacContext, base_url: str) -> dict[str, Any]:
    """Get all the collections available in the given context

    These include a union collection representing all the datasets,
    as well as an individual named collection per dataset.

    :param ctx: a datasets context
    :param base_url: the base URL of the current server
    :return: a STAC dictionary listing the available collections
    """
    return {
        "collections": [_get_datasets_collection(ctx, base_url)]
        + [
            _get_single_dataset_collection(ctx, base_url, c['Identifier'])
            for c in ctx.datasets_ctx.get_dataset_configs()
        ],
        "links": [
            _root_link(base_url),
            {
                "rel": "self",
                "type": "application/json",
                "href": f"{base_url}{PATH_PREFIX}/collections"
            },
            {"rel": "parent", "href": f"{base_url}{PATH_PREFIX}"}
        ]
    }


def get_collection(
    ctx: StacContext, base_url: str, collection_id: str
) -> dict:
    """Return a STAC representation of a collection

    :param ctx: a datasets context
    :param base_url: the base URL of the current server
    :param collection_id: the ID of the collection to describe
    :return: a STAC object representing the collection, if found
    :raises: ApiError.NotFound if no collection with the given ID exists
    """
    ds_ctx = ctx.datasets_ctx
    all_datasets_collection_id, _, _ = _get_collection_metadata(ctx.config)
    collection_ids = [c['Identifier'] for c in ds_ctx.get_dataset_configs()]
    if collection_id in collection_ids:
        return _get_single_dataset_collection(
            ctx, base_url, collection_id, full=True
        )
    elif collection_id == all_datasets_collection_id:
        return _get_datasets_collection(ctx, base_url, full=True)
    else:
        raise ApiError.NotFound(f'Collection "{collection_id}" not found')


def get_single_collection_items(
    ctx: DatasetsContext, base_url: str, collection_id: str
) -> dict:
    """Get the singleton item list for a single-dataset collection

    :param ctx: a datasets context
    :param base_url: the base URL of the current server
    :param collection_id: the ID of a single-dataset collection
    :return: a FeatureCollection dictionary with a singleton feature
        list containing a feature for the requested dataset
    """
    feature = _get_dataset_feature(
        ctx,
        base_url,
        collection_id,
        collection_id,
        DEFAULT_FEATURE_ID,
        full=False,
    )
    self_href = f"{base_url}{PATH_PREFIX}/collections/{collection_id}/items"
    return {
        'type': 'FeatureCollection',
        'features': [feature],
        'links': [
            _root_link(base_url),
            {'rel': 'self', 'type': 'application/json', 'href': self_href},
        ],
        'timeStamp': datetime.datetime.now().astimezone().isoformat()
    }


def get_datasets_collection_items(
    ctx: DatasetsContext,
    base_url: str,
    collection_id: str,
    limit: int = 100,
    cursor: int = 0,
) -> dict:
    """Get the items in the unified datasets collection

    :param ctx: a datasets context
    :param base_url: base URL of the current server
    :param collection_id: the ID of the unified datasets collection
    :param limit: the maximum number of items to return
    :param cursor: the index of the first item to return
    :return: A STAC dictionary of the items in the unified datasets collection,
        limited by the specified limit and cursor values
    """
    _assert_valid_collection(ctx, collection_id)
    all_configs = ctx.get_dataset_configs()
    configs = all_configs[cursor: (cursor + limit)]
    features = []
    for dataset_config in configs:
        dataset_id = dataset_config["Identifier"]
        feature = _get_dataset_feature(
            ctx, base_url, dataset_id, collection_id, dataset_id, full=False
        )
        features.append(feature)
    self_href = f"{base_url}{PATH_PREFIX}/collections/{collection_id}/items"
    links = [
        _root_link(base_url),
        {"rel": "self", "type": "application/json", "href": self_href},
    ]
    if cursor + limit < len(all_configs):
        links.append(
            {
                'rel': 'next',
                'href': self_href + f'?cursor={cursor + limit}&limit={limit}',
            }
        )
    if cursor > 0:
        new_cursor = cursor - limit
        if new_cursor < 0:
            new_cursor = 0
        cursor_param = 'cursor={new_cursor}&' if new_cursor > 0 else ''
        links.append(
            {
                'rel': 'previous',
                'href': self_href + f'?{cursor_param}limit={limit}',
            }
        )
    return {
        "type": "FeatureCollection",
        "features": features,
        "timeStamp": _utc_now(),
        "numberMatched": len(features),
        "numberReturned": len(features),
        "links": links,
    }


def get_collection_item(
    ctx: DatasetsContext, base_url: str, collection_id: str, feature_id: str
) -> dict:
    """Get a specified item from a specified collection

    It is expected that either the collection ID will be that of the
    unified dataset collection and the feature ID will be the dataset ID,
    or that the collection ID will be the dataset ID and the feature ID
    will be the default feature ID for a single-collection dataset.

    :param ctx: a datasets context
    :param base_url: the base URL of the current server
    :param collection_id: the ID of the unified datasets collection or of
        a single-dataset collection
    :param feature_id: the ID of a single dataset within the unified
       collection or of the default feature within a single-dataset collection
    :return: a STAC object representing the specified item, if found
    :raises: ApiError.NotFound, if the specified item is not found
    """
    dataset_ids = {c['Identifier'] for c in ctx.get_dataset_configs()}

    feature_not_found = ApiError.NotFound(
        f'Feature "{feature_id}" not found in collection {collection_id}.'
    )
    if collection_id == DEFAULT_COLLECTION_ID:
        if feature_id in dataset_ids:
            return _get_dataset_feature(
                ctx, base_url, feature_id, collection_id, feature_id, full=True
            )
        else:
            raise feature_not_found
    elif collection_id in dataset_ids:
        if feature_id == DEFAULT_FEATURE_ID:
            return _get_dataset_feature(
                ctx,
                base_url,
                collection_id,
                collection_id,
                feature_id,
                full=True,
            )
        else:
            raise feature_not_found
    else:
        raise ApiError.NotFound(f'Collection "{collection_id}" not found.')


def get_collection_queryables(
    ctx: DatasetsContext, collection_id: str
) -> dict:
    """Get a JSON schema of queryable parameters for the specified collection

    :param ctx: a datasets context
    :param collection_id: the ID of a collection
    :return: a JSON schema of queryable parameters, if the collection was found
    :raises: ApiError.NotFound, if the collection was not found
    """
    _assert_valid_collection(ctx, collection_id)
    schema = JsonObjectSchema(
        title=collection_id, properties={}, additional_properties=False
    )
    return schema.to_dict()


def get_collection_schema(
    ctx: DatasetsContext, base_url: str, collection_id: str
) -> dict:
    """Return a JSON schema for a dataset's data variables

    See links in
    https://docs.ogc.org/DRAFTS/19-087.html#_collection_schema_response_collectionscollectionidschema
    for links to a metaschema defining the schema.

    :param ctx: a datasets context
    :param base_url: the base URL at which this API is being served
    :param collection_id: the ID of a dataset in the provided context
    :return: a JSON schema representing the specified dataset's data variables
    """
    if collection_id == DEFAULT_COLLECTION_ID:
        # The default collection contains multiple datasets, so a range
        # schema doesn't make sense.
        raise ValueError(f'Invalid collection ID {DEFAULT_COLLECTION_ID}')
    _assert_valid_collection(ctx, collection_id)

    ml_dataset = ctx.get_ml_dataset(collection_id)
    ds = ml_dataset.base_dataset

    def get_title(var_name: str) -> str:
        attrs = ds[var_name].attrs
        return attrs['long_name'] if 'long_name' in attrs else var_name

    return {
        '$schema': _JSON_SCHEMA_METASCHEMA,
        '$id': f'{base_url}{PATH_PREFIX}/{collection_id}/schema',
        'title': ds.attrs['title'] if 'title' in ds.attrs else collection_id,
        'type': 'object',
        'properties': {
            var_name: {
                'title': get_title(var_name),
                'type': 'number',
                'x-ogc-property-seq': index + 1,
            } for index, var_name in enumerate(
                # Exclude 0-dimensional vars (usually grid mapping variables)
                {k: v for k, v in ds.data_vars.items() if v.dims != ()}.keys()
            )
        }
    }


# noinspection PyUnusedLocal
def search(ctx: DatasetsContext, base_url: str):
    # TODO: implement me!
    return {}


# noinspection PyUnusedLocal
def _get_datasets_collection(
    ctx: StacContext, base_url: str, full: bool = False
) -> dict:
    ds_ctx = ctx.datasets_ctx
    c_id, c_title, c_description = _get_collection_metadata(ds_ctx.config)
    return {
        "stac_version": STAC_VERSION,
        "stac_extensions": STAC_EXTENSIONS,
        "id": c_id,
        "type": "Collection",
        "title": c_title,
        "description": c_description,
        "license": "proprietary",
        "keywords": [],
        "providers": [],
        "extent": {
            # TODO Replace these placeholder spatial / temporal extents
            # with extents calculated from the datasets.
            "spatial": {"bbox": [[-180.0, -90.0, 180.0, 90.0]]},
            "temporal": {"interval": [["2000-01-01T00:00:00Z", None]]}
        },
        "summaries": {},
        "links": [
            _root_link(base_url),
            {
                "rel": "self",
                "type": "application/json",
                "href": f"{base_url}{PATH_PREFIX}/collections/{c_id}",
                "title": "this collection"
            },
            {
                "rel": "parent",
                "href": f"{base_url}{PATH_PREFIX}/collections",
                "title": "collections list"
            },
            {
                "rel": "items",
                "href": f"{base_url}{PATH_PREFIX}/collections/{c_id}/items",
                "title": "feature collection of data cube items"
            }
        ] + [
            {
                'rel': 'item',
                'href': f'{base_url}{PATH_PREFIX}/collections/'
                        f'{DEFAULT_COLLECTION_ID}/items/{dataset_id}',
                'type': 'application/geo+json',
                'title': f'Feature for the dataset "{dataset_id}"'
            }
            for dataset_id in map(
                lambda c: c['Identifier'], ds_ctx.get_dataset_configs()
            )
        ]
    }


def _get_single_dataset_collection(
    ctx: StacContext, base_url: str, dataset_id: str, full: bool = False
) -> dict:
    ds_ctx = ctx.datasets_ctx
    ml_dataset = ds_ctx.get_ml_dataset(dataset_id)
    dataset = ml_dataset.base_dataset
    grid_bbox = GridBbox(ml_dataset.grid_mapping)
    time_properties = _get_time_properties(dataset)
    if {'start_datetime', 'end_datetime'}.issubset(time_properties):
        time_interval = [
            time_properties['start_datetime'],
            time_properties['end_datetime']
        ]
    else:
        time_interval = [
            time_properties['datetime'],
            time_properties['datetime']
        ]
    storage_crs = crs_to_uri_or_wkt(get_crs_from_dataset(dataset))
    available_crss = [
        crs_to_uri_or_wkt(pyproj.CRS(crs_specifier))
        for crs_specifier in ctx.available_crss
    ]
    if storage_crs not in available_crss:
        available_crss.append(storage_crs)
    gm = GridMapping.from_dataset(dataset)
    result = {
        'assets': _get_assets(ds_ctx, base_url, dataset_id),
        'description': dataset_id,
        'extent': {
            'spatial': {
                'bbox': grid_bbox.as_bbox(),
                'grid': [
                    {'cellsCount': gm.size[0], 'resolution': gm.xy_res[0]},
                    {'cellsCount': gm.size[1], 'resolution': gm.xy_res[1]}
                ]
            },
            'temporal': {
                'interval': [time_interval],
                'grid': get_time_grid(dataset)
            }
        },
        'id': dataset_id,
        'keywords': [],
        'license': 'proprietary',
        'links': [
            _root_link(base_url),
            {
                'rel': 'self',
                'type': 'application/json',
                'href': f'{base_url}{PATH_PREFIX}/collections/{dataset_id}',
                'title': 'this collection'
            },
            {
                'rel': 'parent',
                'href': f'{base_url}{PATH_PREFIX}/collections',
                'title': 'collections list'
            },
            {
                'rel': 'items',
                'href': f'{base_url}{PATH_PREFIX}/collections/'
                        f'{dataset_id}/items',
                'title': 'feature collection of data cube items'
            },
            {
                'rel': 'item',
                'href': f'{base_url}{PATH_PREFIX}/collections/'
                        f'{dataset_id}/items/{DEFAULT_FEATURE_ID}',
                'type': 'application/geo+json',
                'title': f'Feature for the dataset "{dataset_id}"'
            },
            {
                'rel': 'http://www.opengis.net/def/rel/ogc/1.0/coverage',
                'href': f'{base_url}{PATH_PREFIX}/collections/'
                        f'{dataset_id}/coverage?f=json',
                'type': 'application/json',
                'title': f'Coverage for the dataset "{dataset_id}" using '
                         f'OGC API – Coverages, as JSON',
            },
            {
                'rel': 'http://www.opengis.net/def/rel/ogc/1.0/coverage',
                'href': f'{base_url}{PATH_PREFIX}/collections/'
                f'{dataset_id}/coverage?f=netcdf',
                'type': 'application/x-netcdf',
                'title': f'Coverage for the dataset "{dataset_id}" using '
                f'OGC API – Coverages, as NetCDF',
            },
            {
                'rel': 'http://www.opengis.net/def/rel/ogc/1.0/coverage',
                'href': f'{base_url}{PATH_PREFIX}/collections/'
                f'{dataset_id}/coverage?f=geotiff',
                'type': 'image/tiff; application=geotiff',
                'title': f'Coverage for the dataset "{dataset_id}" using '
                f'OGC API – Coverages, as GeoTIFF',
            },
            {
                'rel': _REL_SCHEMA,
                'href': f'{base_url}{PATH_PREFIX}/collections/'
                        f'{dataset_id}/schema?f=json',
                'type': 'application/json',
                'title': 'Schema (as JSON)',
            },
            {
                'rel': _REL_RANGETYPE,
                'href': f'{base_url}{PATH_PREFIX}/collections/'
                        f'{dataset_id}/coverage/rangetype?f=json',
                'type': 'application/json',
                'title': 'Range type of the coverage',
            },
            {
                'rel': _REL_DOMAINSET,
                'href': f'{base_url}{PATH_PREFIX}/collections/'
                        f'{dataset_id}/coverage/domainset?f=json',
                'type': 'application/json',
                'title': 'Domain set of the coverage',
            },
        ],
        'providers': [],
        'stac_version': STAC_VERSION,
        'summaries': {},
        'title': dataset_id,
        'type': 'Collection',
        'storageCRS': storage_crs,
        'crs': available_crss
    }
    result.update(_get_cube_properties(ds_ctx, dataset_id))
    return result


def crs_to_uri_or_wkt(crs: pyproj.CRS) -> str:
    auth_and_code = crs.to_authority()
    if auth_and_code is not None:
        authority, code = auth_and_code
        version = 0  # per https://docs.ogc.org/pol/09-048r6.html#toc13
        return (
            f'http://www.opengis.net/def/crs/' f'{authority}/{version}/{code}'
        )
    else:
        return crs.to_wkt()


class GridBbox:
    """Utility class to transform and manipulate bounding box data"""

    def __init__(self, grid_mapping: GridMapping):
        transformer = pyproj.Transformer.from_crs(
            grid_mapping.crs, CRS_CRS84, always_xy=True
        )
        bbox = grid_mapping.xy_bbox
        (self.x1, self.x2), (self.y1, self.y2) = transformer.transform(
            (bbox[0], bbox[2]), (bbox[1], bbox[3])
        )

    def as_bbox(self) -> list:
        return [self.x1, self.y1, self.x2, self.y2]

    def as_geometry(self) -> dict[str, Union[str, list]]:
        return {
            "type": "Polygon",
            "coordinates": [
                [
                    [self.x1, self.y1],
                    [self.x1, self.y2],
                    [self.x2, self.y2],
                    [self.x2, self.y1],
                    [self.x1, self.y1],
                ]
            ],
        }


def get_time_grid(ds: xr.Dataset) -> dict[str, Any]:
    """Return a dictionary representing the grid for a dataset's time variable

    The dictionary format is defined by the schema at
    https://github.com/opengeospatial/ogcapi-coverages/blob/master/standard/openapi/schemas/common-geodata/extent.yaml

    :param ds: a dataset
    :return: a dictionary representation of the grid of the dataset's
             time variable
    """
    if 'time' not in ds:
        return {}

    if ds.dims['time'] < 2:
        time_is_regular = False
    else:
        time_diffs = ds.time.diff(dim='time').astype('uint64')
        time_is_regular = np.allclose(time_diffs[0], time_diffs)

    return dict([
        ('cellsCount', ds.dims['time']),
        ('resolution',
         pd.Timedelta((ds.time[1] - ds.time[0]).values).isoformat())
        if time_is_regular else
        ('coordinates', [pd.Timestamp(t.values).isoformat() for t in ds.time])
    ])


# noinspection PyUnusedLocal
def _get_dataset_feature(
    ctx: DatasetsContext,
    base_url: str,
    dataset_id: str,
    collection_id: str,
    feature_id: str,
    full: bool = False,
) -> dict:
    bbox = GridBbox(ctx.get_ml_dataset(dataset_id).grid_mapping)

    return {
        "stac_version": STAC_VERSION,
        "stac_extensions": STAC_EXTENSIONS,
        "type": "Feature",
        "id": feature_id,
        "bbox": bbox.as_bbox(),
        "geometry": bbox.as_geometry(),
        "properties": _get_cube_properties(ctx, dataset_id),
        "collection": collection_id,
        "links": [
            _root_link(base_url),
            {
                "rel": "self",
                "href": f"{base_url}{PATH_PREFIX}/collections/{collection_id}"
                        f"/items/{feature_id}"
            },
            {
                "rel": "collection",
                "href": f"{base_url}{PATH_PREFIX}/collections/{collection_id}"
            },
            {
                "rel": "parent",
                "href": f"{base_url}{PATH_PREFIX}/collections/{collection_id}"
            }
        ],
        "assets": _get_assets(ctx, base_url, dataset_id)
    }


def _get_cube_properties(ctx: DatasetsContext, dataset_id: str):
    ml_dataset = ctx.get_ml_dataset(dataset_id)
    grid_mapping = ml_dataset.grid_mapping
    dataset = ml_dataset.base_dataset

    cube_dimensions = get_datacube_dimensions(dataset, grid_mapping)

    return {
        "cube:dimensions": cube_dimensions,
        "cube:variables": _get_dc_variables(dataset, cube_dimensions),
        "xcube:dims": to_json_value(dataset.dims),
        "xcube:data_vars": _get_xc_variables(dataset.data_vars),
        "xcube:coords": _get_xc_variables(dataset.coords),
        "xcube:attrs": to_json_value(dataset.attrs),
        **(_get_time_properties(dataset)),
    }


def _get_assets(ctx: DatasetsContext, base_url: str, dataset_id: str):
    ml_dataset = ctx.get_ml_dataset(dataset_id)
    dataset = ml_dataset.base_dataset
    xcube_data_vars = _get_xc_variables(dataset.data_vars)
    first_var_name = next(iter(xcube_data_vars))["name"]
    first_var = dataset[first_var_name]
    first_var_extra_dims = first_var.dims[0:-2]

    thumbnail_query = ''
    if first_var_extra_dims:
        thumbnail_query_params = []
        for dim in first_var_extra_dims:
            val = 0
            if dim in dataset:
                coord = dataset[dim]
                if coord.ndim == 1 and coord.size > 0:
                    val = coord[0].to_numpy()
            thumbnail_query_params.append(f'{dim}={val}')
        thumbnail_query = '?' + '&'.join(thumbnail_query_params)

    tiles_query = ''
    if first_var_extra_dims:
        tiles_query = '?' + '&'.join(
            ['%s=<%s>' % (d, d) for d in first_var_extra_dims]
        )

    # TODO: Prefer original storage location.
    #       The "s3" operation is default.
    default_storage_url = f"{base_url}/s3/datasets"

    return {
        "analytic": {
            "title": f"{dataset_id} data access",
            "roles": ["data"],
            "type": "application/zarr",
            "href": f"{default_storage_url}/{dataset_id}.zarr",
            "xcube:analytic": {
                v['name']: {
                    "title": f"{v['name']} data access",
                    "roles": ["data"],
                    "type": "application/zarr",
                    "href": f"{default_storage_url}/"
                            f"{dataset_id}.zarr/{v['name']}"
                }
                for v in xcube_data_vars
            }
        },
        "visual": {
            "title": f"{dataset_id} visualisation",
            "roles": ["visual"],
            "type": "image/png",
            "href": (
                f"{base_url}/tiles/{dataset_id}/<variable>"
                + "/{z}/{y}/{x}"
                + tiles_query
            ),
            "xcube:visual": {
                v['name']: {
                    "title": f"{v['name']} visualisation",
                    "roles": ["visual"],
                    "type": "image/png",
                    "href": (
                        f"{base_url}/tiles/{dataset_id}/{v['name']}"
                        + "/{z}/{y}/{x}"
                        + tiles_query
                    ),
                }
                for v in xcube_data_vars
            },
        },
        "thumbnail": {
            "title": f"{dataset_id} thumbnail",
            "roles": ["thumbnail"],
            "type": "image/png",
            "href": f"{base_url}/tiles/{dataset_id}/{first_var_name}"
                    f"/0/0/0{thumbnail_query}"
        }
    }


def _get_time_properties(dataset):
    if 'time' in dataset:
        time_var = dataset['time']
        start_time = to_json_value(time_var[0])
        end_time = to_json_value(time_var[-1])
        time_properties = (
            {'datetime': start_time}
            if start_time == end_time
            else {
                'datetime': None,
                'start_datetime': start_time,
                'end_datetime': end_time
            }
        )
    else:
        time_properties = {
            # TODO Decide what to use as a fall-back datetime
            'datetime': '2000-01-01T00:00:00Z'
        }
    return time_properties


def _get_xc_variables(
    variables: Mapping[Hashable, xr.DataArray]
) -> List[Dict[str, Any]]:
    """Create the value of the "xcube:coords" or
    "xcube:data_vars" property for the given *dataset*.
    """
    return [
        _get_xc_variable(var_name, var) for var_name, var in variables.items()
    ]


def _get_xc_variable(var_name: Hashable, var: xr.DataArray) -> Dict[str, Any]:
    """Create an entry of the value of the "xcube:coords" or
    "xcube:data_vars" property for the given *dataset*.
    """
    return {
        "name": str(var_name),
        "dtype": str(var.dtype),
        "dims": to_json_value(var.dims),
        "chunks": to_json_value(var.chunks) if var.chunks else None,
        "shape": to_json_value(var.shape),
        "attrs": to_json_value(var.attrs),
        # "encoding": to_json_value(var.encoding),
    }


def get_datacube_dimensions(
    dataset: xr.Dataset, grid_mapping: GridMapping
) -> Dict[str, Any]:
    """Create the value of the "datacube:dimensions" property
    for the given *dataset*.

    :param dataset: the dataset to describe
    :param grid_mapping: the dataset's grid mapping
    :return: a dictionary of the datacube properties of the dataset
    """
    x_dim_name, y_dim_name = grid_mapping.xy_dim_names
    x_var_name, y_var_name = grid_mapping.xy_var_names
    dc_dimensions = {
        x_dim_name: _get_dc_spatial_dimension(
            dataset[x_var_name], "x", grid_mapping
        ),
        y_dim_name: _get_dc_spatial_dimension(
            dataset[y_var_name], "y", grid_mapping
        ),
    }
    if (
        "time" in dataset.dims
        and "time" in dataset.coords
        and dataset["time"].ndim == 1
    ):
        dc_dimensions.update(time=_get_dc_temporal_dimension(dataset["time"]))
    for dim_name in dataset.dims.keys():
        if (
            dim_name not in {x_dim_name, y_dim_name, "time"}
            and dim_name in dataset
        ):
            dc_dimensions.update(
                {dim_name: _get_dc_additional_dimension(dataset[dim_name])}
            )
    return dc_dimensions


def _get_dc_spatial_dimension(
    var: xr.DataArray, axis: str, grid_mapping: GridMapping
) -> Dict[str, Any]:
    """Create a spatial dimension of the "datacube:dimensions" property
    for the given *var* and *axis*.
    """
    asset = _get_dc_dimension(var, "spatial", axis=axis)
    if axis == "x":
        extent = grid_mapping.x_min, grid_mapping.x_max
        step = grid_mapping.x_res if grid_mapping.is_regular else None
    else:
        extent = grid_mapping.y_min, grid_mapping.y_max
        step = grid_mapping.y_res if grid_mapping.is_regular else None
    asset["extent"] = to_json_value(extent)
    asset["step"] = to_json_value(step)
    asset["reference_system"] = to_json_value(grid_mapping.crs.srs)
    return asset


def _get_dc_temporal_dimension(var: xr.DataArray) -> Dict[str, Any]:
    """Create a temporal dimension of the "datacube:dimensions" property
    for the given time *var*.
    """
    asset = _get_dc_dimension(var, "temporal", axis=None, drop_unit=True)
    asset["values"] = [to_json_value(t) for t in var.values]
    return asset


def _get_dc_additional_dimension(
    var: xr.DataArray, type_: str = "unknown"
) -> Dict[str, Any]:
    """Create an additional dimension of the "datacube:dimensions" property
    for the given *var* and *type*.
    """
    asset = _get_dc_dimension(var, type_, axis=None)
    if var.ndim == 1:
        asset["range"] = [to_json_value(var[0]), to_json_value(var[-1])]
        if var.size > 1:
            diff_var = np.diff(var)
            if np.issubdtype(var.dtype, np.number) and np.allclose(
                np.diff(diff_var), 0
            ):
                asset["step"] = to_json_value(diff_var[0])
        if "step" not in asset and var.size < _MAX_NUM_VALUES:
            asset["values"] = [to_json_value(t) for t in var.values]
    return asset


def _get_dc_dimension(
    var: xr.DataArray,
    type_: str,
    axis: Optional[str] = None,
    drop_unit: bool = False,
) -> Dict[str, Any]:
    """Create a generic dimension of the "datacube:dimensions" property
    for the given *var*, *type*, and optional *axis*.
    """
    asset = dict(type=type_)
    if axis is not None:
        asset.update(axis=axis)
    _set_dc_description(asset, var)
    if not drop_unit:
        _set_dc_unit(asset, var)
    return asset


def _get_dc_variables(dataset: xr.Dataset, dc_dimensions):
    """Create the value of the "datacube:variables" property
    for the given *dataset*.
    """
    return dict(
        **__get_dc_variables(dataset.data_vars, "data", dc_dimensions),
        **__get_dc_variables(dataset.coords, "auxiliary", dc_dimensions),
    )


def __get_dc_variables(
    variables: Mapping[Hashable, xr.DataArray],
    type: str,
    dc_dimensions: Dict[str, Any],
):
    """Create a partial value of the "datacube:variables" property
    for the given *variables* and *type*.
    """
    return {
        str(var_name): _get_dc_variable(var, type)
        for var_name, var in variables.items()
        if var_name not in dc_dimensions and var.ndim >= 1
    }


def _get_dc_variable(var: xr.DataArray, type: str) -> Dict[str, Any]:
    """Create a generic variable of the "datacube:variables" property
    for the given *var*, *type*, and optional *axis*.
    """
    asset = dict(type=type, dimensions=list(var.dims))
    _set_dc_description(asset, var)
    _set_dc_unit(asset, var)
    return asset


def _set_dc_description(asset, var):
    """Set the "description" property of given asset, if any."""
    description = _get_str_attr(
        var.attrs, ['description', 'title', 'long_name']
    )
    if description:
        asset.update(description=description)


def _set_dc_unit(asset, var):
    """Set the "unit" property of given asset, if any."""
    unit = _get_str_attr(var.attrs, ['unit', 'units'])
    if unit:
        asset.update(unit=unit)


def _get_str_attr(attrs: Dict[str, Any], keys: List[str]) -> Optional[str]:
    for k in keys:
        v = attrs.get(k)
        if isinstance(v, str) and v:
            return v
    return None


def _assert_valid_collection(ctx: DatasetsContext, collection_id: str):
    # c_id, _, _ = _get_collection_metadata(ctx.config)
    collection_ids = [c['Identifier'] for c in ctx.get_dataset_configs()]
    if (
        collection_id not in collection_ids
        and collection_id != DEFAULT_COLLECTION_ID
    ):
        raise ApiError.NotFound(f'Collection "{collection_id}" not found')


def _get_catalog_metadata(config: ServerConfig):
    stac_config = config.get("STAC", {})
    catalog_id = stac_config.get("Identifier", DEFAULT_CATALOG_ID)
    catalog_title = stac_config.get("Title", DEFAULT_CATALOG_TITLE)
    catalog_description = stac_config.get(
        "Description", DEFAULT_CATALOG_DESCRIPTION
    )
    return catalog_id, catalog_title, catalog_description


def _get_collection_metadata(config: ServerConfig):
    stac_config = config.get("STAC", {})
    collection_config = stac_config.get("Collection", {})
    collection_id = collection_config.get("Identifier", DEFAULT_COLLECTION_ID)
    collection_title = collection_config.get("Title", DEFAULT_COLLECTION_TITLE)
    collection_description = collection_config.get(
        "Description", DEFAULT_COLLECTION_DESCRIPTION
    )
    return collection_id, collection_title, collection_description


def _utc_now():
    return datetime.datetime.utcnow().replace(microsecond=0).isoformat() + 'Z'


class CollectionNotFoundException(Exception):
    pass


class DimensionNotFoundException(Exception):
    pass



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import cProfile
import glob
import io
import os
import pstats
import sys
import time
import traceback
import warnings
from typing import Any, Callable, Dict, Sequence, Tuple

import xarray as xr

from xcube.core.dsio import DatasetIO, find_dataset_io, guess_dataset_format, rimraf
from xcube.core.evaluate import evaluate_dataset
from xcube.core.gen.defaults import DEFAULT_OUTPUT_PATH, DEFAULT_OUTPUT_RESAMPLING, DEFAULT_OUTPUT_SIZE
from xcube.core.gen.iproc import InputProcessor, find_input_processor_class
from xcube.core.gridmapping import GridMapping
from xcube.core.gridmapping import CRS_WGS84
from xcube.core.optimize import optimize_dataset
from xcube.core.select import select_spatial_subset, select_variables_subset
from xcube.core.timecoord import add_time_coords, from_time_in_days_since_1970
from xcube.core.timeslice import find_time_slice
from xcube.core.update import update_dataset_attrs, update_dataset_temporal_attrs, update_dataset_var_attrs
from xcube.util.config import NameAnyDict, NameDictPairList, to_resolved_name_dict_pairs


def gen_cube(input_paths: Sequence[str] = None,
             input_processor_name: str = None,
             input_processor_params: Dict = None,
             input_reader_name: str = None,
             input_reader_params: Dict[str, Any] = None,
             output_region: Tuple[float, float, float, float] = None,
             output_size: Tuple[int, int] = DEFAULT_OUTPUT_SIZE,
             output_resampling: str = DEFAULT_OUTPUT_RESAMPLING,
             output_path: str = DEFAULT_OUTPUT_PATH,
             output_writer_name: str = None,
             output_writer_params: Dict[str, Any] = None,
             output_metadata: NameAnyDict = None,
             output_variables: NameDictPairList = None,
             processed_variables: NameDictPairList = None,
             profile_mode: bool = False,
             no_sort_mode: bool = False,
             append_mode: bool = None,
             dry_run: bool = False,
             monitor: Callable[..., None] = None) -> bool:
    """
    Generate a xcube dataset from one or more input files.

    :param no_sort_mode:
    :param input_paths: The input paths.
    :param input_processor_name: Name of a registered input processor
        (xcube.core.gen.inputprocessor.InputProcessor) to be used to transform the inputs.
    :param input_processor_params: Parameters to be passed to the input processor.
    :param input_reader_name: Name of a registered input reader (xcube.core.util.dsio.DatasetIO).
    :param input_reader_params: Parameters passed to the input reader.
    :param output_region: Output region as tuple of floats: (lon_min, lat_min, lon_max, lat_max).
    :param output_size: The spatial dimensions of the output as tuple of ints: (width, height).
    :param output_resampling: The resampling method for the output.
    :param output_path: The output directory.
    :param output_writer_name: Name of an output writer
        (xcube.core.util.dsio.DatasetIO) used to write the cube.
    :param output_writer_params: Parameters passed to the output writer.
    :param output_metadata: Extra metadata passed to output cube.
    :param output_variables: Output variables.
    :param processed_variables: Processed variables computed on-the-fly.
    :param profile_mode: Whether profiling should be enabled.
    :param append_mode: Deprecated. The function will always either insert, replace, or append new time slices.
    :param dry_run: Doesn't write any data. For testing.
    :param monitor: A progress monitor.
    :return: True for success.
    """

    if append_mode is not None:
        warnings.warn('append_mode in gen_cube() is deprecated, '
                      'time slices will now always be inserted, replaced, or appended.')

    if input_processor_name is None:
        input_processor_name = 'default'
    elif input_processor_name == '':
        raise ValueError('input_processor_name must not be empty')

    input_processor_class = find_input_processor_class(input_processor_name)
    if not input_processor_class:
        raise ValueError(f'Unknown input_processor_name {input_processor_name!r}')

    if not issubclass(input_processor_class, InputProcessor):
        raise ValueError(f'Invalid input_processor_name {input_processor_name!r}: '
                         f'must name a sub-class of {InputProcessor.__qualname__}')

    try:
        input_processor = input_processor_class(**(input_processor_params or {}))
    except (ValueError, TypeError) as e:
        raise ValueError(f'Invalid input_processor_name or input_processor_params: {e}') from e

    input_reader = find_dataset_io(input_reader_name or input_processor.input_reader)
    if not input_reader:
        raise ValueError(f'Unknown input_reader_name {input_reader_name!r}')

    if not output_path:
        raise ValueError('Missing output_path')

    output_writer_name = output_writer_name or guess_dataset_format(output_path)
    if not output_writer_name:
        raise ValueError(f'Failed to guess output_writer_name from path {output_path}')
    output_writer = find_dataset_io(output_writer_name, modes={'w', 'a'})
    if not output_writer:
        raise ValueError(f'Unknown output_writer_name {output_writer_name!r}')

    if monitor is None:
        # noinspection PyUnusedLocal
        def monitor(*args):
            pass

    input_paths = [input_file for f in input_paths for input_file in glob.glob(f, recursive=True)]

    effective_input_reader_params = dict(input_processor.input_reader_params or {})
    effective_input_reader_params.update(input_reader_params or {})

    if not no_sort_mode and len(input_paths) > 1:
        input_paths = _get_sorted_input_paths(input_processor, input_reader, effective_input_reader_params, input_paths)

    if not dry_run:
        output_dir = os.path.abspath(os.path.dirname(output_path))
        os.makedirs(output_dir, exist_ok=True)

    effective_output_writer_params = output_writer_params or {}

    status = False

    ds_count = len(input_paths)
    ds_count_ok = 0
    ds_index = 0
    for input_file in input_paths:
        monitor(f'processing dataset {ds_index + 1} of {ds_count}: {input_file!r}...')
        # noinspection PyTypeChecker
        status = _process_input(input_processor,
                                input_reader,
                                effective_input_reader_params,
                                output_writer,
                                effective_output_writer_params,
                                input_file,
                                output_size,
                                output_region,
                                output_resampling,
                                output_path,
                                output_metadata,
                                output_variables,
                                processed_variables,
                                profile_mode,
                                dry_run,
                                monitor)
        ds_index += 1
        if status:
            ds_count_ok += 1

    monitor(f'{ds_count_ok} of {ds_count} datasets processed successfully, '
            f'{ds_count - ds_count_ok} were dropped due to errors')

    return status


def _process_input(input_processor: InputProcessor,
                   input_reader: DatasetIO,
                   input_reader_params: Dict[str, Any],
                   output_writer: DatasetIO,
                   output_writer_params: Dict[str, Any],
                   input_file: str,
                   output_size: Tuple[int, int],
                   output_region: Tuple[float, float, float, float],
                   output_resampling: str,
                   output_path: str,
                   output_metadata: NameAnyDict = None,
                   output_variables: NameDictPairList = None,
                   processed_variables: NameDictPairList = None,
                   profile_mode: bool = False,
                   dry_run: bool = False,
                   monitor: Callable[..., None] = None) -> bool:
    monitor('reading input slice...')
    # noinspection PyBroadException
    try:
        input_dataset = input_reader.read(input_file, **input_reader_params)
        monitor(f'Dataset read:\n{input_dataset}')
    except Exception as e:
        monitor(f'Error: cannot read input: {e}: skipping...')
        traceback.print_exc(file=sys.stderr)
        return False

    time_range = input_processor.get_time_range(input_dataset)
    if time_range[0] > time_range[1]:
        monitor('Error: start time is greater than end time: skipping...')
        return False

    if output_variables:
        output_variables = to_resolved_name_dict_pairs(output_variables, input_dataset, keep=True)
    else:
        output_variables = [(var_name, None) for var_name in input_dataset.data_vars]

    time_index, update_mode = find_time_slice(output_path,
                                              from_time_in_days_since_1970((time_range[0] + time_range[1]) / 2))

    width, height = output_size
    x_min, y_min, x_max, y_max = output_region
    xy_res = max((x_max - x_min) / width, (y_max - y_min) / height)
    tile_size = _get_tile_size(output_writer_params)

    output_geom = GridMapping.regular(size=output_size,
                                      xy_min=(x_min, y_min),
                                      xy_res=xy_res,
                                      crs=CRS_WGS84,
                                      tile_size=tile_size)

    steps = []

    # noinspection PyShadowingNames
    def step1(input_slice):
        return input_processor.pre_process(input_slice)

    steps.append((step1, 'pre-processing input slice'))

    grid_mapping = None

    # noinspection PyShadowingNames
    def step1a(input_slice):
        nonlocal grid_mapping
        grid_mapping = GridMapping.from_dataset(input_slice)
        subset = select_spatial_subset(input_slice,
                                       xy_bbox=output_geom.xy_bbox,
                                       xy_border=output_geom.x_res,
                                       ij_border=1,
                                       grid_mapping=grid_mapping)
        if subset is None:
            monitor('no spatial overlap with input')
        elif subset is not input_slice:
            grid_mapping = GridMapping.from_dataset(subset)
        return subset

    steps.append((step1a, 'spatial subsetting'))

    # noinspection PyShadowingNames
    def step2(input_slice):
        return evaluate_dataset(input_slice, processed_variables=processed_variables)

    steps.append((step2, 'computing input slice variables'))

    # noinspection PyShadowingNames
    def step3(input_slice):
        extra_vars = input_processor.get_extra_vars(input_slice)
        selected_variables = set([var_name for var_name, _ in output_variables])
        selected_variables.update(extra_vars or set())
        return select_variables_subset(input_slice, selected_variables)

    steps.append((step3, 'selecting input slice variables'))

    # noinspection PyShadowingNames
    def step4(input_slice):
        # noinspection PyTypeChecker
        return input_processor.process(input_slice,
                                       geo_coding=grid_mapping,
                                       output_geom=output_geom,
                                       output_resampling=output_resampling,
                                       include_non_spatial_vars=False)

    steps.append((step4, 'transforming input slice'))

    if time_range is not None:
        def step5(input_slice):
            return add_time_coords(input_slice, time_range)

        steps.append((step5, 'adding time coordinates to input slice'))

    def step6(input_slice):
        return update_dataset_var_attrs(input_slice, output_variables)

    steps.append((step6, 'updating variable attributes of input slice'))

    def step7(input_slice):
        return input_processor.post_process(input_slice)

    steps.append((step7, 'post-processing input slice'))

    if update_mode == 'create':
        def step8(input_slice):
            if not dry_run:
                rimraf(output_path)
                output_writer.write(input_slice, output_path, **output_writer_params)
                _update_cube(output_writer, output_path, global_attrs=output_metadata, temporal_only=False)
            return input_slice

        steps.append((step8, f'creating input slice in {output_path}'))

    elif update_mode == 'append':
        def step8(input_slice):
            if not dry_run:
                output_writer.append(input_slice, output_path, **output_writer_params)
                _update_cube(output_writer, output_path, temporal_only=True)
            return input_slice

        steps.append((step8, f'appending input slice to {output_path}'))

    elif update_mode == 'insert':
        def step8(input_slice):
            if not dry_run:
                output_writer.insert(input_slice, time_index, output_path)
                _update_cube(output_writer, output_path, temporal_only=True)
            return input_slice

        steps.append((step8, f'inserting input slice before index {time_index} in {output_path}'))

    elif update_mode == 'replace':
        def step8(input_slice):
            if not dry_run:
                output_writer.replace(input_slice, time_index, output_path)
                _update_cube(output_writer, output_path, temporal_only=True)
            return input_slice

        steps.append((step8, f'replacing input slice at index {time_index} in {output_path}'))

    if profile_mode:
        pr = cProfile.Profile()
        pr.enable()

    status = True
    try:
        num_steps = len(steps)
        dataset = input_dataset
        total_t1 = time.perf_counter()
        for step_index in range(num_steps):
            transform, label = steps[step_index]
            step_t1 = time.perf_counter()
            monitor(f'step {step_index + 1} of {num_steps}: {label}...')
            dataset = transform(dataset)
            step_t2 = time.perf_counter()
            if dataset is None:
                monitor(f'  {label} terminated after {step_t2 - step_t1} seconds, skipping input slice')
                status = False
                break
            monitor(f'  {label} completed in {step_t2 - step_t1} seconds')
        total_t2 = time.perf_counter()
        monitor(f'{num_steps} steps took {total_t2 - total_t1} seconds to complete')
    except RuntimeError as e:
        monitor(f'Error: something went wrong during processing, skipping input slice: {e}')
        traceback.print_exc(file=sys.stderr)
        status = False
    finally:
        input_dataset.close()

    if profile_mode:
        # noinspection PyUnboundLocalVariable
        pr.disable()
        s = io.StringIO()
        ps = pstats.Stats(pr, stream=s).sort_stats('cumtime')
        ps.print_stats()
        monitor(s.getvalue())

    return status


def _get_tile_size(output_writer_params):
    tile_size = None
    if 'chunksizes' in output_writer_params:
        chunksizes = output_writer_params['chunksizes']
        if 'lon' in chunksizes or 'lat' in chunksizes:
            tile_size = chunksizes.get('lon', 512), chunksizes.get('lat', 512)
        elif 'x' in chunksizes or 'y' in chunksizes:
            tile_size = chunksizes.get('x', 512), chunksizes.get('y', 512)
    return tile_size


def _update_cube(output_writer: DatasetIO,
                 output_path: str,
                 global_attrs: Dict = None,
                 temporal_only: bool = False):
    cube = output_writer.read(output_path)
    if temporal_only:
        cube = update_dataset_temporal_attrs(cube,
                                             update_existing=True, in_place=True)
    else:
        cube = update_dataset_attrs(cube,
                                    update_existing=True, in_place=True)
    cube_attrs = dict(cube.attrs)
    cube.close()

    if global_attrs:
        cube_attrs.update(global_attrs)

    output_writer.update(output_path, global_attrs=cube_attrs)


def _get_sorted_input_paths(input_processor, input_reader: DatasetIO, input_reader_params: Dict[str, Any], input_paths: Sequence[str]):
    input_path_list = []
    time_list = []
    for input_file in input_paths:
#        with xr.open_dataset(input_file) as dataset:
        with input_reader.read(input_file, **input_reader_params) as dataset:
            t1, t2 = input_processor.get_time_range(dataset)
            time_list.append((t1 + t2) / 2)
            input_path_list.append(input_file)
            tuple_seq = zip(time_list, input_path_list)
    input_paths = [e[1] for e in sorted(tuple_seq, key=lambda e: e[0])]
    return input_paths



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import concurrent.futures
import inspect
import os.path
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Tuple, Dict, Type, Sequence, \
    Generic, TypeVar, Union, Callable, Awaitable, Mapping

from .asyncexec import AsyncExecution
from ..util.assertions import assert_instance
from ..util.assertions import assert_true
from ..util.frozen import FrozenDict
from ..util.jsonschema import JsonObjectSchema

_SERVER_CONTEXT_ATTR_NAME = '__xcube_server_context'
_HTTP_METHODS = {'head', 'get', 'post', 'put', 'delete', 'options'}

ArgT = TypeVar('ArgT')
ReturnT = TypeVar('ReturnT')
# API Context type variable
ServerContextT = TypeVar("ServerContextT", bound="Context")
ApiContextT = TypeVar("ApiContextT", bound="Context")

JSON = Union[
    None,
    bool,
    int,
    float,
    str,
    List["JSON"],
    Dict[str, "JSON"],
]

ServerConfigObject = FrozenDict[str, Any]
ServerConfig = FrozenDict[str, Any]

builtin_type = type


class Api(Generic[ServerContextT]):
    """
    A server API.

    The most common purpose of this class is to
    add a new API to the server by the means of routes.

    Every may produce API context objects for a given server
    configuration.

    If the server configuration changes, the API is asked to
    create a new context object.

    However, an API may be just programmatic and serve as a
    web server middleware. It can then still provide
    the context for other dependent APIs.

    May be derived by clients to override the methods

    * `create_ctx` - to create an API-specific context object;
    * `on_start` - to do things on server start;
    * `on_stop` - to do things on server stop.

    Note that these methods can be also be effectively
    implemented by respectively passing the *create_ctx*, *on_start*,
    *on_stop* arguments to the constructor.

    Each extension API module must export an instance of this
    class. A typical use case of this class:

    ```
    class DatasetsContext(ApiContext)
        def update(self, prev_ctx: Optional[Context]):
            config = self.config
            ...

        def get_datasets(self):
            ...

    api = Api("datasets",
              config_schema=DATASET_CONFIG_SCHEMA,
              create_ctx=DatasetsApiContext)

    @api.route("/datasets")
    class DatasetsHandler(ApiHandler[DatasetsContext]):
        def get(self):
            return self.ctx.get_datasets()
    ```

    :param name: The API name. Must be unique within a server.
    :param version: The API version. Defaults to "0.0.0".
    :param routes: Optional list of initial routes.
        A route is a tuple of the form (route-pattern, handler-class) or
        (route-pattern, handler-class, handler-kwargs). The handler-class
        must be derived from ApiHandler.
    :param required_apis: Sequence of names of other required APIs.
    :param optional_apis: Sequence of names of other optional APIs.
    :param config_schema: Optional JSON schema for the API's configuration.
        If not given, or None is passed, the API is assumed to
        have no configuration.
    :param create_ctx: Optional API context factory.
        If given, must be a callable that accepts the server context
        and returns a ``Context`` instance for the API.
        Called when a new context is required after configuration changes.
    :param on_start: Optional start handler.
        If given, must be a callable that accepts the server context.
        Called when the server starts.
    :param on_stop: Optional stop handler.
        If given, must be a callable that accepts the server context.
        Called when the server stopped.
    """

    def __init__(
            self,
            name: str, /,
            version: str = '0.0.0',
            description: Optional[str] = None,
            routes: Optional[Sequence["ApiRoute"]] = None,
            static_routes: Optional[Sequence["ApiStaticRoute"]] = None,
            required_apis: Optional[Sequence[str]] = None,
            optional_apis: Optional[Sequence[str]] = None,
            config_schema: Optional[JsonObjectSchema] = None,
            create_ctx: Optional[
                Callable[["Context"], Optional[ServerContextT]]
            ] = None,
            on_start: Optional[
                Callable[["Context"], Any]
            ] = None,
            on_stop: Optional[
                Callable[["Context"], Any]
            ] = None,
    ):
        assert_instance(name, str, 'name')
        assert_instance(version, str, 'version')
        if description is not None:
            assert_instance(description, str, 'description')
        if config_schema is not None:
            assert_instance(config_schema, JsonObjectSchema, 'config_schema')
        if on_start is not None:
            assert_true(callable(on_start),
                        message='on_start must be callable')
        if on_stop is not None:
            assert_true(callable(on_stop),
                        message='on_stop must be callable')
        self._name = name
        self._version = version
        self._description = description
        self._required_apis = tuple(required_apis or ())
        self._optional_apis = tuple(optional_apis or ())
        self._routes: List[ApiRoute] = list(routes or [])
        self._static_routes: List[ApiStaticRoute] = list(static_routes or [])
        self._config_schema = config_schema
        self._create_ctx = create_ctx or ApiContext
        self._on_start = on_start or self._handle_event
        self._on_stop = on_stop or self._handle_event

    @property
    def name(self) -> str:
        """The name of this API."""
        return self._name

    @property
    def version(self) -> str:
        """The version of this API."""
        return self._version

    @property
    def description(self) -> Optional[str]:
        """The description of this API."""
        return self._description or (getattr(self, '__doc__', None)
                                     if self.__class__ is not Api else None)

    @property
    def required_apis(self) -> Tuple[str]:
        """The names of other required APIs."""
        return self._required_apis

    @property
    def optional_apis(self) -> Tuple[str]:
        """The names of other optional APIs."""
        return self._optional_apis

    def static_route(self,
                     path: str,
                     default_filename: Optional[str] = None,
                     **openapi_metadata):
        """
        Decorator that adds static route to this API.

        The decorator target must be a function that returns the route's
        local root directory path either as a string or a ``pathlib.Path``.
        If it can not determine an existing path, it should return None.

        :param path: The route path.
        :param default_filename: Optional default filename,
            e.g., "index.html".
        :param openapi_metadata: Optional OpenAPI GET operation metadata.
        """

        def decorator_func(get_root_path: Callable[[], Optional[str]]):
            root_path = get_root_path()
            if root_path is not None:
                self._static_routes.append(
                    ApiStaticRoute(path,
                                   str(root_path),
                                   api_name=self.name,
                                   default_filename=default_filename,
                                   openapi_metadata=openapi_metadata)
                )

        return decorator_func

    def route(self, path: str, **handler_kwargs):
        """
        Decorator that adds a route to this API.

        The decorator target must be a class derived from ApiHandler.

        :param path: The route path.
        :param handler_kwargs: Optional keyword arguments passed to
            ApiHandler constructor.
        :return: A decorator function that receives a
            class derived from ApiHandler
        """

        def decorator_func(handler_cls: Type[ApiHandler]):
            self._routes.append(ApiRoute(self.name,
                                         path,
                                         handler_cls,
                                         handler_kwargs))
            return handler_cls

        return decorator_func

    def operation(self,
                  operation_id: Optional[str] = None,
                  summary: Optional[str] = None,
                  description: Optional[str] = None,
                  parameters: Optional[List[Dict[str, Any]]] = None,
                  request_body: Optional[Dict[str, Any]] = None,
                  responses: Optional[Dict[str, Dict]] = None,
                  tags: Optional[str] = None,
                  **kwargs):
        """
        Decorator that adds OpenAPI 3.0 information to an
        API handler's operation,
        i.e. one of the get, post, put, delete, or options methods.

        :param operation_id: a string identifier for the operation
        :param summary: a summary of the operation
        :param description: A description of the operation. CommonMark syntax
                            may be used.
        :param parameters: List of dictionaries, one per parameter, defining
                           OpenAPI Parameter objects
        :param request_body: A dictionary defining an OpenAPI Request Body
                             object
        :param responses: A dictionary defining an OpenAPI Responses object
        :param tags: OpenAPI tags
        :param kwargs: OpenAPI parameters

        :return: A decorator function that receives and returns an
                 ApiHandler's operation.
        """
        openapi = {
            "operationId": operation_id or kwargs.pop("operationId", None),
            "summary": summary,
            "description": description,
            "parameters": parameters,
            "requestBody": request_body or kwargs.pop("requestBody", None),
            "responses": responses,
            "tags": tags,
            **kwargs
        }
        openapi = {k: v for k, v in openapi.items() if v is not None}

        def decorator_func(target: Union[Type[ApiHandler], Callable]):
            if inspect.isfunction(target) \
                    and hasattr(target, '__name__') \
                    and target.__name__ in _HTTP_METHODS:
                setattr(target, "__openapi__", openapi)
            else:
                raise TypeError(f'API {self.name}:'
                                f' @operation() decorator'
                                f' must be used with one of the'
                                f' HTTP methods of an {ApiHandler.__name__}')
            return target

        return decorator_func

    @property
    def static_routes(self) -> Tuple["ApiStaticRoute"]:
        """The static routes provided by this API."""
        return tuple(self._static_routes)

    @property
    def routes(self) -> Tuple["ApiRoute"]:
        """The routes provided by this API."""
        return tuple(self._routes)

    @property
    def config_schema(self) -> Optional[JsonObjectSchema]:
        """Get the JSON schema for the configuration of this API."""
        return self._config_schema

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def create_ctx(self, server_ctx: "Context") -> ServerContextT:
        """Create a new context object for this API.
        The default implementation uses the *create_ctx*
        argument passed to the constructor, if any,
        to instantiate an API context using *ctx* as only argument.
        Otherwise, a new instance of ``ApiContext`` is returned.
        Should not be called directly.

        :param server_ctx: The server's current context.
        :return: An instance of ``Context``
        """
        return self._create_ctx(server_ctx)

    def on_start(self, server_ctx: "Context"):
        """Called when the server is started.
        Can be overridden to initialize the API.
        Should not be called directly.

        The default implementation calls the *on_start*
        argument passed to the constructor, if any.

        :param server_ctx: The server's current context
        """
        return self._on_start(server_ctx)

    def on_stop(self, server_ctx: "Context"):
        """Called when the server is stopped.
        Can be overridden to initialize the API.
        Should not be called directly.

        The default implementation calls the *on_stop*
        argument passed to the constructor, if any.

        :param server_ctx: The server's current context
        """
        return self._on_stop(server_ctx)

    def endpoints(self) -> list[dict[str, Union[str, list[str]]]]:
        """Describe endpoints provided by this API

        Returns a list of dictionaries of the form
        {'path': '/some/path', 'methods': ['get', 'post']}

        :return: a list of dictionaries, each describing an endpoint
        """
        return [route.path_and_methods() for route in self.routes]

    def _handle_event(self, server_ctx: "Context"):
        """Do nothing."""

    def __repr__(self):
        return f'Api({self.name!r}, version={self.version!r})'


class Context(AsyncExecution, ABC):
    """The interface for server context objects."""

    @property
    @abstractmethod
    def apis(self) -> Tuple[Api]:
        """The APIs used by the server."""

    @abstractmethod
    def get_open_api_doc(self, include_all: bool = False) -> Dict[str, Any]:
        """The OpenAPI JSON document for the server."""

    @property
    @abstractmethod
    def config(self) -> ServerConfig:
        """The server's current configuration."""

    @abstractmethod
    def get_api_ctx(self,
                    api_name: str,
                    cls: Optional[Type[ApiContextT]] = None) \
            -> Optional[ApiContextT]:
        """
        Get the API context for *api_name*.
        Can be used to access context objects of other APIs.

        The keyword argument *cls* can be used to assert a specific
        type if of context.

        :param api_name: The name of a registered API.
        :param cls: Optional context class.
            If given, must be a class derived from `ApiContext`.
        :return: The API context object for *api_name*,
            or None if no such exists.
        """

    @abstractmethod
    def on_update(self, prev_context: Optional["Context"]):
        """Called when the server configuration changed.
        Must be implemented by derived classes in order to update
        this context with respect to the current configuration
        ``self.config`` and the given *prev_context*, if any.
        The method shall not be called directly.

        :param prev_context: The previous context instance.
            Will be ``None`` if ``on_update()`` is called for the
            very first time.
        """

    @abstractmethod
    def on_dispose(self):
        """Called if this context will never be used again.
        May be overridden by derived classes in order to
        dispose allocated resources.
        The default implementation does nothing.
        The method shall not be called directly.
        """


class ApiContext(Context):
    """
    An implementation of the server context to be used by APIs.

    A typical use case is to cache computationally expensive
    resources served by a particular API.

    An instance of this class is created for every API unless
    an API provides its own, specific context.

    Specific context classes should derive from ``ApiContext`` and

    * may override the `on_update()` method in order
      to initialize or update this context object state with
      respect to the current server configuration, or with
      respect to other API context object states.
    * may override the `on_dispose()` method to empty any caches
      and close access to resources.
    * must call the super class constructor with the
      *server_ctx* context, from their own constructor, if any.

    :param server_ctx: The server context.
    """

    def __init__(self, server_ctx: Context):
        self._server_ctx = server_ctx

    @property
    def server_ctx(self) -> Context:
        """The server context."""
        return self._server_ctx

    @property
    def apis(self) -> Tuple[Api]:
        """Return the server context's ``apis`` property."""
        return self.server_ctx.apis

    def get_open_api_doc(self, include_all: bool = False) -> Dict[str, Any]:
        """Return the server context's ``apis`` property."""
        return self.server_ctx.get_open_api_doc(include_all=include_all)

    @property
    def config(self) -> ServerConfig:
        """Return the server context's ``config`` property."""
        return self.server_ctx.config

    def get_api_ctx(self,
                    api_name: str,
                    cls: Optional[Type[ApiContextT]] = None) \
            -> Optional[ApiContextT]:
        """Calls the server context's ``get_api_ctx()`` method."""
        return self.server_ctx.get_api_ctx(api_name, cls=cls)

    def call_later(self,
                   delay: Union[int, float],
                   callback: Callable,
                   *args,
                   **kwargs) -> object:
        """Calls the server context's ``call_later()`` method."""
        return self.server_ctx.call_later(delay, callback,
                                          *args, **kwargs)

    def run_in_executor(self,
                        executor: Optional[concurrent.futures.Executor],
                        function: Callable[..., ReturnT],
                        *args: Any,
                        **kwargs: Any) -> Awaitable[ReturnT]:
        """Calls the server context's ``run_in_executor()`` method."""
        return self.server_ctx.run_in_executor(executor, function,
                                               *args, **kwargs)

    def on_update(self, prev_context: Optional["Context"]):
        """Does nothing."""

    def on_dispose(self):
        """Does nothing."""


class ApiRequest:

    @abstractmethod
    def url_for_path(self,
                     path: str,
                     query: Optional[str] = None,
                     reverse: bool = False) -> str:
        """Get the reverse URL for the given *path* component and
        optional *query* string.

        The *reverse* flag controls if configuration parameter
        ``reverse_url_prefix`` shall be used for prefixing
        the returned URL.

        :param path: The path component.
        :param query: Optional query string.
        :param reverse: Whether the reverse URL shall be returned.
        :return: Full reverse URL for *path* and *query*.
        """

    @property
    def base_url(self) -> str:
        """Full base URL for this request, without query."""
        return self.url_for_path("", reverse=False)

    @property
    def reverse_base_url(self) -> str:
        """Full reverse base URL for this request, without query."""
        return self.url_for_path("", reverse=True)

    @property
    @abstractmethod
    def url(self) -> str:
        """The full, original URL of the request.
        Includes query parameters, if any.
        """

    @property
    @abstractmethod
    def headers(self) -> Mapping[str, str]:
        """The request headers."""

    @property
    @abstractmethod
    def body(self) -> bytes:
        """The request body."""

    @property
    @abstractmethod
    def json(self) -> JSON:
        """The request body as JSON value."""

    @abstractmethod
    def make_query_lower_case(self):
        """Make the query keys case-insensitive."""

    @property
    @abstractmethod
    def query(self) -> Mapping[str, Sequence[str]]:
        """The request query arguments."""

    # noinspection PyShadowingBuiltins
    def get_query_arg(self,
                      name: str,
                      type: Optional[Type[ArgT]] = None,
                      default: Optional[ArgT] = None) -> Optional[ArgT]:
        """Get the value of query argument given by *name*. To force
        conversion to a specific target data type use the optional *type*
        argument. If *type* is not given, but *default* is, then *type*
        will be inferred from *default*.

        :param name: The name of the argument
        :param type: The requested data type.
            Must be a callable type, e.g. bool, int.
        :param default: Optional default value.
        :return: The value of the query argument.
        """
        if type is None and default is not None:
            type = builtin_type(default)
            type = type if callable(type) else None
        values = self.get_query_args(name, type=type)
        return values[0] if values else default

    # noinspection PyShadowingBuiltins
    @abstractmethod
    def get_query_args(self,
                       name: str,
                       type: Optional[Type[ArgT]] = None) -> Sequence[ArgT]:
        """Get the values of query argument given by *name*.
        If *type* is given, a sequence of that type will be returned.
        :param name: The name of the argument
        :param type: The requested data type.
            Must be a callable type, e.g. bool, int.
        :return: The values of the query argument.
        """


class ApiResponse(ABC):
    @abstractmethod
    def set_status(self, status_code: int, reason: Optional[str] = None):
        """Set the HTTP status code and optionally the reason."""

    @abstractmethod
    def set_header(self, name: str, value: str):
        """Set the HTTP header *name* to given *value*."""

    @abstractmethod
    def write(self, 
              data: Union[str, bytes, JSON],
              content_type: Optional[str] = None):
        """Write data."""

    @abstractmethod
    def finish(self, 
               data: Union[str, bytes, JSON] = None,
               content_type: Optional[str] = None):
        """Finish the response (and submit it)."""


class ApiHandler(Generic[ServerContextT], ABC):
    """
    Base class for all API handlers.

    :param ctx: The API context.
    :param request: The API handler's request.
    :param response: The API handler's response.
    :param kwargs: Client keyword arguments (not used in base class).
    """

    def __init__(self,
                 ctx: Context,
                 request: ApiRequest,
                 response: ApiResponse,
                 **kwargs: Any):
        assert_true(not kwargs,
                    message=f"Unknown keyword(s) passed to"
                            f" {self.__class__.__name__}:"
                            f" {', '.join(kwargs.keys())}.")
        self._ctx = ctx
        self._request = request
        self._response = response

    @property
    def ctx(self) -> ServerContextT:
        """The API's context object."""
        # noinspection PyTypeChecker
        return self._ctx

    @property
    def request(self) -> ApiRequest:
        """The request that provides the handler's input."""
        return self._request

    @property
    def response(self) -> ApiResponse:
        """The response that provides the handler's output."""
        return self._response

    @classmethod
    def defined_methods(cls) -> list[str]:
        """List the methods handled by this handler

        Any method with an __openapi__ attribute is assumed to be
        handled.

        :return: a list of the method names handled by this handler
        """
        return [
            name
            for name in ['head', 'get', 'post', 'put', 'delete', 'options']
            if hasattr(getattr(cls, name), '__openapi__')
        ]

    # HTTP methods

    def head(self, *args, **kwargs):
        raise ApiError.MethodNotAllowed("method HEAD not allowed")

    def get(self, *args, **kwargs):
        raise ApiError.MethodNotAllowed("method GET not allowed")

    def post(self, *args, **kwargs):
        raise ApiError.MethodNotAllowed("method POST not allowed")

    def put(self, *args, **kwargs):
        raise ApiError.MethodNotAllowed("method PUT not allowed")

    def delete(self, *args, **kwargs):
        raise ApiError.MethodNotAllowed("method DELETE not allowed")

    def options(self, *args, **kwargs):
        # Warning, naive implementation:
        # By default, we allow for pre-flight OPTIONS requests.
        # We could improve by returning 204 only for the method given by
        # header "Access-Control-Request-Method", which effectively
        # has an implementation in a client's ApiHandler.
        # For more, see
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/OPTIONS
        self.response.set_status(204)
        self.response.finish()


class ApiRoute:
    """
    An API route.

    :param api_name: The name of the API to which this route belongs to.
    :param path: The route path which may include path variable templates.
    :param handler_cls: The route handler class.
        Must be derived from ```ApiHandler```.
    :param handler_kwargs: Optional keyword arguments passed to
        the *handler_cls* when it is instantiated.
    """

    def __init__(self,
                 api_name: str,
                 path: str,
                 handler_cls: Type[ApiHandler],
                 handler_kwargs: Optional[Dict[str, Any]] = None):
        assert_instance(api_name, str, name="api_name")
        assert_instance(path, str, name="path")
        assert_instance(handler_cls, type, name="handler_cls")
        assert_true(issubclass(handler_cls, ApiHandler),
                    message=f'handler_cls must be a subclass'
                            f' of {ApiHandler.__name__},'
                            f' was {handler_cls}')
        assert_instance(handler_kwargs, (type(None), dict),
                        name="handler_kwargs")
        self.api_name = api_name
        self.path = path
        self.handler_cls = handler_cls
        self.handler_kwargs = dict(handler_kwargs or {})

    def __eq__(self, other) -> bool:
        if isinstance(other, ApiRoute):
            return self.api_name == other.api_name \
                   and self.path == other.path \
                   and self.handler_cls == other.handler_cls \
                   and self.handler_kwargs == other.handler_kwargs
        return False

    def __hash__(self) -> int:
        return hash(self.api_name) \
               + 2 * hash(self.path) \
               + 4 * hash(self.handler_cls) \
               + 16 * hash(tuple(sorted(tuple(self.handler_kwargs.items()),
                                        key=lambda p: p[0])))

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        args = (f"{self.api_name!r},"
                f" {self.path!r},"
                f" {self.handler_cls.__name__}")
        if self.handler_kwargs:
            args += f", handler_kwargs={self.handler_kwargs!r}"
        return f"ApiRoute({args})"

    def path_and_methods(self) -> dict[str, Union[str, list[str]]]:
        """Describe the path and methods of this route

        Returns a dictionary of the form
        {'path': '/this/routes/path', 'methods': ['get', 'post']}

        :return: a dictionary describing this route
        """
        return {
            'path': self.path,
            'methods': self.handler_cls.defined_methods()
        }


class ApiStaticRoute:
    """
    An API static route.

    :param path: The route path.
    :param dir_path: A local directory path.
    :param default_filename: Optional default filename, e.g., "index.html".
    :param api_name: Optional name of the API to which this route belongs to.
    :param openapi_metadata: Optional OpenAPI operation metadata.
    """

    def __init__(self,
                 path: str,
                 dir_path: str,
                 default_filename: Optional[str] = None,
                 api_name: Optional[str] = None,
                 openapi_metadata: Optional[Dict[str, Any]] = None):
        assert_instance(path, str, name="path")
        assert_instance(dir_path, str, name="dir_path")
        assert_true(os.path.abspath(dir_path),
                    message="dir_path must be an absolute local path")
        assert_instance(default_filename, (type(None), str),
                        name="default_filename")
        assert_instance(api_name, (type(None), str),
                        name="api_name")
        assert_instance(openapi_metadata, (type(None), dict),
                        name="openapi_metadata")
        self.path = path
        self.dir_path = dir_path
        self.default_filename = default_filename
        self.api_name = api_name
        self.openapi_metadata = openapi_metadata


class ApiError(Exception):
    """
    An API error.
    This exception should be raised to terminate the current request
    with a defined HTTP status code.

    :param status_code: The HTTP status code
    :param message: Optional message
    """

    def __init__(self,
                 status_code: int,
                 message: Optional[str] = None):
        super().__init__(status_code, message)

    BadRequest: Type["_DerivedApiError"]
    Unauthorized: Type["_DerivedApiError"]
    Forbidden: Type["_DerivedApiError"]
    NotFound: Type["_DerivedApiError"]
    MethodNotAllowed: Type["_DerivedApiError"]
    Conflict: Type["_DerivedApiError"]
    Gone: Type["_DerivedApiError"]
    ContentTooLarge: Type["_DerivedApiError"]
    InternalServerError: Type["_DerivedApiError"]
    NotImplemented: Type["_DerivedApiError"]
    InvalidServerConfig: Type["_DerivedApiError"]
    UnsupportedMediaType: Type["_DerivedApiError"]

    @property
    def status_code(self) -> int:
        return self.args[0]

    @property
    def message(self) -> Optional[str]:
        return self.args[1]

    def __str__(self):
        text = f'HTTP status {self.status_code}'
        if self.message:
            text += f': {self.message}'
        return text


class _DerivedApiError(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(500, message=message)


class _BadRequest(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(400, message=message)


class _Unauthorized(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(401, message=message)


class _Forbidden(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(403, message=message)


class _NotFound(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(404, message=message)


class _MethodNotAllowed(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(405, message=message)


class _Conflict(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(409, message=message)


class _Gone(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(410, message=message)


class _ContentTooLarge(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(413, message=message)


class _UnsupportedMediaType(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(415, message=message)


class _InternalServerError(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(500, message=message)


class _NotImplemented(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(501, message=message)


class _InvalidServerConfig(ApiError):
    def __init__(self, message: Optional[str] = None):
        super().__init__(580, message=message)


ApiError.BadRequest = _BadRequest
ApiError.Unauthorized = _Unauthorized
ApiError.Forbidden = _Forbidden
ApiError.NotFound = _NotFound
ApiError.MethodNotAllowed = _MethodNotAllowed
ApiError.UnsupportedMediaType = _UnsupportedMediaType
ApiError.Conflict = _Conflict
ApiError.Gone = _Gone
ApiError.InternalServerError = _InternalServerError
ApiError.NotImplemented = _NotImplemented
ApiError.InvalidServerConfig = _InvalidServerConfig
ApiError.ContentTooLarge = _ContentTooLarge



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


import fnmatch
import itertools
import os
import os.path
import warnings
from functools import cached_property
from typing import Any, Dict, List, Optional, Tuple, Callable, Collection, \
    Set, Mapping, Union

import numpy as np
import pandas as pd
import pyproj
import xarray as xr

from xcube.constants import LOG
from xcube.core.mldataset import BaseMultiLevelDataset, \
    IdentityMultiLevelDataset
from xcube.core.mldataset import MultiLevelDataset
from xcube.core.mldataset import augment_ml_dataset
from xcube.core.mldataset import open_ml_dataset_from_python_code
from xcube.core.store import DATASET_TYPE
from xcube.core.store import DataStoreConfig
from xcube.core.store import DataStorePool
from xcube.core.store import DatasetDescriptor
from xcube.core.store import MULTI_LEVEL_DATASET_TYPE
from xcube.core.tile import get_var_cmap_params
from xcube.core.tile import get_var_valid_range
from xcube.server.api import Context, ApiError
from xcube.server.api import ServerConfig
from xcube.server.config import is_absolute_path
from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.cache import parse_mem_size
from xcube.util.cmaps import ColormapRegistry
from xcube.util.cmaps import load_custom_colormap
from xcube.webapi.common.context import ResourcesContext
from xcube.webapi.places import PlacesContext

COMPUTE_DATASET = 'compute_dataset'
COMPUTE_VARIABLES = 'compute_variables'

# We use tilde, because it is not a reserved URI characters
STORE_DS_ID_SEPARATOR = '~'
FS_TYPE_TO_PROTOCOL = {
    'local': 'file',
    'obs': 's3',
    'file': 'file',
    's3': 's3',
    'memory': 'memory'
}
NON_MEMORY_FILE_SYSTEMS = ['local', 'obs', 'file', 's3']

ALL_PLACES = "all"

MultiLevelDatasetOpener = Callable[
    ["DatasetsContext", ServerConfig],
    MultiLevelDataset
]

DatasetConfig = Mapping[str, Any]


class DatasetsContext(ResourcesContext):

    def __init__(self,
                 server_ctx: Context,
                 ml_dataset_openers=None):
        super().__init__(server_ctx)
        # noinspection PyTypeChecker
        self._places_ctx: PlacesContext = server_ctx.get_api_ctx("places")
        assert isinstance(self._places_ctx, PlacesContext)
        self._ml_dataset_openers = ml_dataset_openers
        # cache for all dataset configs
        # contains tuples of form (MultiLevelDataset, dataset_config)
        self._dataset_cache = dict()
        self._data_store_pool, self._dataset_configs = \
            self._process_dataset_configs(self.config, self.base_dir)
        self._cm_styles, self._colormap_registry = self._get_cm_styles()

    def on_dispose(self):
        with self.rlock:
            # Close all datasets
            for ml_dataset, _ in self._dataset_cache.values():
                # noinspection PyBroadException
                try:
                    ml_dataset.close()
                except Exception:
                    pass
            # Clear all caches
            if self._dataset_cache:
                self._dataset_cache.clear()
            if self._data_store_pool:
                self._data_store_pool.remove_all_store_configs()
            self._dataset_configs = None

    @property
    def places_ctx(self):
        return self._places_ctx

    @property
    def dataset_cache(self) \
            -> Dict[str, Tuple[MultiLevelDataset, DatasetConfig]]:
        return self._dataset_cache

    @cached_property
    def access_control(self) -> Dict[str, Any]:
        return self.config.get('AccessControl', {})

    @cached_property
    def required_scopes(self) -> List[str]:
        return self.access_control.get('RequiredScopes', [])

    @property
    def colormap_registry(self) -> ColormapRegistry:
        return self._colormap_registry

    def get_required_dataset_scopes(
            self,
            dataset_config: DatasetConfig
    ) -> Set[str]:
        return self._get_required_scopes(dataset_config,
                                         'read:dataset', 'Dataset',
                                         dataset_config['Identifier'])

    def get_required_variable_scopes(
            self,
            dataset_config: DatasetConfig,
            var_name: str
    ) -> Set[str]:
        return self._get_required_scopes(dataset_config,
                                         'read:variable', 'Variable',
                                         var_name)

    def _get_required_scopes(self,
                             dataset_config: DatasetConfig,
                             base_scope: str,
                             value_name: str,
                             value: str) -> Set[str]:
        required_global_scopes = set(self.required_scopes)
        required_dataset_scopes = set(
            dataset_config.get('AccessControl', {}).get('RequiredScopes', [])
        )
        if not required_global_scopes and not required_dataset_scopes:
            return required_global_scopes
        required_dataset_scopes = required_global_scopes \
            .union(required_dataset_scopes)
        base_scope_prefix = base_scope + ':'
        required_dataset_scopes = {
            scope
            for scope in required_dataset_scopes
            if scope == base_scope or scope.startswith(base_scope_prefix)
        }
        pattern_scope = base_scope_prefix + '{' + value_name + '}'
        if pattern_scope in required_dataset_scopes:
            # Replace "{base_scope}:{value_name}" by "{base_scope}:{value}"
            required_dataset_scopes.remove(pattern_scope)
            required_dataset_scopes.add(base_scope_prefix + value)
        return required_dataset_scopes

    def get_ml_dataset(self, ds_id: str) -> MultiLevelDataset:
        ml_dataset, _ = self._get_dataset_entry(ds_id)
        return ml_dataset

    def set_ml_dataset(self, ml_dataset: MultiLevelDataset):
        self._set_dataset_entry((ml_dataset,
                                 dict(Identifier=ml_dataset.ds_id,
                                      Hidden=True)))

    def add_dataset(
            self,
            dataset: Union[xr.Dataset, MultiLevelDataset],
            ds_id: Optional[str] = None,
            title: Optional[str] = None,
            style: Optional[str] = None,
            color_mappings: Dict[str, Dict[str, Any]] = None
    ) -> str:
        """Add a dataset to the configuration
        """
        assert_instance(dataset, (xr.Dataset, MultiLevelDataset), 'dataset')
        if isinstance(dataset, xr.Dataset):
            ml_dataset = BaseMultiLevelDataset(dataset, ds_id=ds_id)
        else:
            ml_dataset = dataset
            if ds_id:
                ml_dataset = IdentityMultiLevelDataset(dataset, ds_id=ds_id)
            dataset = ml_dataset.base_dataset
        ds_id = ml_dataset.ds_id
        dataset_configs = self.get_dataset_configs()
        for index, dataset_config in enumerate(dataset_configs):
            if dataset_config["Identifier"] == ds_id:
                del dataset_configs[index]
                break
        dataset_config = dict(Identifier=ds_id,
                              Title=title or dataset.attrs.get("title",
                                                               ds_id))
        if style is not None:
            dataset_config.update(dict(Style=style))
        if color_mappings is not None:
            style = style or ds_id
            dataset_config.update(dict(Style=style))
            self._cm_styles[style] = color_mappings
        self._dataset_cache[ds_id] = ml_dataset, dataset_config
        self._dataset_configs.append(dataset_config)
        return ds_id

    def remove_dataset(self, ds_id: str):
        assert_instance(ds_id, str, 'ds_id')
        assert_given(ds_id, 'ds_id')
        if ds_id in self._dataset_cache:
            del self._dataset_cache[ds_id]
        self._dataset_configs = [
            dc
            for dc in self._dataset_configs
            if dc["Identifier"] != ds_id
        ]

    def add_ml_dataset(self,
                       ml_dataset: MultiLevelDataset,
                       ds_id: Optional[str] = None,
                       title: Optional[str] = None):
        if ds_id:
            ml_dataset.ds_id = ds_id
        else:
            ds_id = ml_dataset.ds_id
        self._set_dataset_entry((ml_dataset,
                                 dict(Identifier=ds_id,
                                      Title=title)))

    def get_dataset(self,
                    ds_id: str,
                    expected_var_names: Collection[str] = None) -> xr.Dataset:
        ml_dataset, _ = self._get_dataset_entry(ds_id)
        dataset = ml_dataset.base_dataset
        if expected_var_names:
            for var_name in expected_var_names:
                if var_name not in dataset:
                    raise ApiError.NotFound(
                        f'Variable "{var_name}" not found'
                        f' in dataset "{ds_id}"'
                    )
        return dataset

    def get_time_series_dataset(self,
                                ds_id: str,
                                var_name: str = None) -> xr.Dataset:
        dataset_config = self.get_dataset_config(ds_id)
        ts_ds_name = dataset_config.get('TimeSeriesDataset', ds_id)
        try:
            # Try to get more efficient, time-chunked dataset
            return self.get_dataset(ts_ds_name,
                                    expected_var_names=[var_name]
                                    if var_name else None)
        except ApiError.NotFound:
            # This happens, if the dataset pointed to by 'TimeSeriesDataset'
            # does not contain the variable given by var_name.
            return self.get_dataset(ds_id,
                                    expected_var_names=[var_name]
                                    if var_name else None)

    def get_variable_for_z(self,
                           ds_id: str,
                           var_name: str,
                           z_index: int) -> xr.DataArray:
        ml_dataset = self.get_ml_dataset(ds_id)
        index = ml_dataset.num_levels - 1 - z_index
        if index < 0 or index >= ml_dataset.num_levels:
            raise ApiError.NotFound(
                f'Variable "{var_name}" has no z-index {z_index}'
                f' in dataset "{ds_id}"'
            )
        dataset = ml_dataset.get_dataset(index)
        if var_name not in dataset:
            raise ApiError.NotFound(
                f'Variable "{var_name}" not found'
                f' in dataset "{ds_id}"'
            )
        return dataset[var_name]

    @classmethod
    def _maybe_assign_store_instance_ids(
            cls,
            dataset_configs: List[Dict[str, Any]],
            data_store_pool: DataStorePool,
            base_dir: str
    ) -> None:
        assignable_dataset_configs = [
            dc for dc in dataset_configs
            if 'StoreInstanceId' not in dc
               and dc.get('FileSystem', 'file') in NON_MEMORY_FILE_SYSTEMS
        ]
        # split into sub-lists according to file system and
        # non-root store params
        config_lists = []
        for config in assignable_dataset_configs:
            store_params = cls._get_other_store_params_than_root(config)
            file_system = config.get('FileSystem', 'file')
            appended = False
            for config_list in config_lists:
                if config_list[0] == file_system and \
                        config_list[1] == store_params:
                    config_list[2].append(config)
                    appended = True
                    break
            if not appended:
                config_lists.append((file_system, store_params, [config]))

        for file_system, store_params, config_list in config_lists:
            # Retrieve paths per configuration
            paths = [dc['Path'] for dc in config_list]
            # Determine common prefixes of paths (and call them roots)
            roots = _get_roots(paths)
            for root in roots:
                fs_protocol = FS_TYPE_TO_PROTOCOL.get(file_system,
                                                      file_system)
                store_params_with_root = store_params.copy()
                if fs_protocol == "file" and not is_absolute_path(root):
                    store_params_with_root['root'] = f"{base_dir}/{root}"
                else:
                    store_params_with_root['root'] = root
                # See if there already is a store with this configuration
                data_store_config = DataStoreConfig(
                    store_id=fs_protocol,
                    store_params=store_params_with_root
                )
                store_instance_id = data_store_pool. \
                    get_store_instance_id(data_store_config)
                if not store_instance_id:
                    # Create new store with new unique store instance id
                    counter = 1
                    while data_store_pool.has_store_instance(
                            f'{fs_protocol}_{counter}'):
                        counter += 1
                    store_instance_id = f'{fs_protocol}_{counter}'
                    data_store_pool.add_store_config(store_instance_id,
                                                     data_store_config)
                for config in config_list:
                    path = config['Path']
                    if path.startswith(root) and \
                            config.get('StoreInstanceId') is None:
                        config['StoreInstanceId'] = store_instance_id
                        new_path = path[len(root):]
                        while new_path.startswith("/") or \
                                new_path.startswith("\\"):
                            new_path = new_path[1:]
                        config['Path'] = new_path

    @classmethod
    def _get_other_store_params_than_root(
            cls, dataset_config: DatasetConfig
    ) -> Dict:
        file_system = FS_TYPE_TO_PROTOCOL.get(
            dataset_config.get('FileSystem', 'file')
        )
        if file_system != 's3':
            return {}
        storage_options = dict()
        if 'Anonymous' in dataset_config:
            storage_options['anon'] = dataset_config['Anonymous']
        client_kwargs = dict(
        )
        if 'Endpoint' in dataset_config:
            client_kwargs['endpoint_url'] = dataset_config['Endpoint']
        if 'Region' in dataset_config:
            client_kwargs['region_name'] = dataset_config['Region']
        storage_options['client_kwargs'] = client_kwargs
        store_params = dict(storage_options=storage_options)
        return store_params

    @classmethod
    def get_dataset_configs_from_stores(
            cls,
            data_store_pool: DataStorePool
    ) -> List[DatasetConfig]:
        all_dataset_configs: List[DatasetConfig] = []
        for store_instance_id in data_store_pool.store_instance_ids:
            LOG.info(f'Scanning store {store_instance_id!r}')
            data_store_config = data_store_pool.get_store_config(
                store_instance_id
            )
            data_store = data_store_pool.get_store(store_instance_id)
            # Note by forman: This iterator chaining is inefficient.
            # Preferably, we should offer
            #
            # store_dataset_ids = data_store.get_data_ids(
            #     data_type=(DATASET_TYPE, MULTI_LEVEL_DATASET_TYPE)
            # )
            #
            store_dataset_ids = itertools.chain(
                data_store.get_data_ids(
                    data_type=DATASET_TYPE
                ),
                data_store.get_data_ids(
                    data_type=MULTI_LEVEL_DATASET_TYPE
                )
            )
            for store_dataset_id in store_dataset_ids:
                dataset_config_base = {}
                store_dataset_configs: List[ServerConfig] \
                    = data_store_config.user_data
                if store_dataset_configs:
                    for store_dataset_config in store_dataset_configs:
                        dataset_id_pattern = store_dataset_config.get(
                            'Path', '*'
                        )
                        if fnmatch.fnmatch(store_dataset_id,
                                           dataset_id_pattern):
                            dataset_config_base = store_dataset_config
                            break
                        else:
                            dataset_config_base = None
                if dataset_config_base is not None:
                    LOG.debug(f'Selected dataset {store_dataset_id!r}')
                    dataset_config = dict(
                        StoreInstanceId=store_instance_id,
                        **dataset_config_base
                    )
                    if dataset_config.get('Identifier') is not None:
                        if dataset_config['Path'] == store_dataset_id:
                            # we will use the preconfigured identifier
                            all_dataset_configs.append(dataset_config)
                            continue
                        raise ApiError.InvalidServerConfig(
                            'User-defined identifiers can only be assigned'
                            ' to datasets with non-wildcard paths.'
                        )
                    dataset_config['Path'] = store_dataset_id
                    dataset_config['Identifier'] = \
                        f'{store_instance_id}{STORE_DS_ID_SEPARATOR}' \
                        f'{store_dataset_id}'
                    all_dataset_configs.append(dataset_config)

        # # Just for testing:
        # debug_file = 'all_dataset_configs.json'
        # with open(debug_file, 'w') as stream:
        #     json.dump(all_dataset_configs, stream)
        #     LOG.debug(f'Wrote file {debug_file!r}')

        return all_dataset_configs

    def new_dataset_metadata(self,
                             store_instance_id: str,
                             dataset_id: str) -> Optional[DatasetDescriptor]:
        data_store = self._data_store_pool.get_store(store_instance_id)
        dataset_metadata = data_store.describe_data(
            dataset_id,
            data_type='dataset'
        )
        if dataset_metadata.crs is not None:
            crs = pyproj.CRS.from_string(dataset_metadata.crs)
            if not crs.is_geographic:
                LOG.warning(f'Ignoring dataset {dataset_id!r} from'
                            f' store instance {store_instance_id!r}'
                            f' because it uses a non-geographic CRS')
                return None
        # noinspection PyTypeChecker
        return dataset_metadata

    def get_dataset_config(self, ds_id: str) -> DatasetConfig:
        dataset_configs = self.get_dataset_configs()
        dataset_config = next(
            (dsd for dsd in dataset_configs if dsd['Identifier'] == ds_id),
            None
        )
        if dataset_config is None:
            raise ApiError.NotFound(f'Dataset "{ds_id}" not found')
        return dataset_config

    def get_dataset_configs(self) -> List[DatasetConfig]:
        assert self._dataset_configs is not None
        return self._dataset_configs

    def get_data_store_pool(self) -> DataStorePool:
        assert self._data_store_pool is not None
        return self._data_store_pool

    @classmethod
    def _process_dataset_configs(
            cls,
            config: ServerConfig,
            base_dir: str
    ) -> Tuple[DataStorePool, List[Dict[str, Any]]]:
        data_store_configs = config.get('DataStores', [])
        dataset_configs = config.get('Datasets', [])

        data_store_pool = DataStorePool()
        for data_store_config_dict in data_store_configs:
            store_instance_id = data_store_config_dict.get('Identifier')
            store_id = data_store_config_dict.get('StoreId')
            store_params = data_store_config_dict.get('StoreParams', {})
            store_dataset_configs = data_store_config_dict.get('Datasets')
            store_config = DataStoreConfig(store_id,
                                           store_params=store_params,
                                           user_data=store_dataset_configs)
            data_store_pool.add_store_config(store_instance_id,
                                             store_config)
        dataset_configs = \
            dataset_configs + cls.get_dataset_configs_from_stores(
                data_store_pool
            )
        # Allow dataset_configs to be writable, because
        # _maybe_assign_store_instance_ids() will change
        # entries:
        dataset_configs = [dict(c) for c in dataset_configs]
        cls._maybe_assign_store_instance_ids(dataset_configs,
                                             data_store_pool,
                                             base_dir)
        return data_store_pool, dataset_configs

    def get_rgb_color_mapping(
            self,
            ds_id: str,
            norm_range: Tuple[float, float] = (0., 1.)
    ) -> Tuple[List[Optional[str]], List[Tuple[float, float]]]:
        var_names = [None, None, None]
        norm_ranges = [norm_range, norm_range, norm_range]
        color_mappings = self.get_color_mappings(ds_id)
        if color_mappings:
            rgb_mapping = color_mappings.get('rgb')
            if rgb_mapping:
                components = 'Red', 'Green', 'Blue'
                for i in range(3):
                    component = components[i]
                    component_config = rgb_mapping.get(component, {})
                    var_name = component_config.get('Variable')
                    norm_vmin, norm_vmax = component_config.get('ValueRange',
                                                                norm_range)
                    var_names[i] = var_name
                    norm_ranges[i] = norm_vmin, norm_vmax
        return var_names, norm_ranges

    def get_color_mapping(self,
                          ds_id: str,
                          var_name: str) -> Tuple[str, Tuple[float, float]]:
        cmap_name = None
        cmap_vmin, cmap_vmax = None, None
        color_mappings = self.get_color_mappings(ds_id)
        if color_mappings:
            color_mapping = color_mappings.get(var_name)
            if color_mapping:
                assert 'ColorFile' not in color_mapping
                cmap_name = color_mapping.get('ColorBar')
                cmap_vmin, cmap_vmax = color_mapping.get('ValueRange',
                                                         (None, None))

        cmap_range = cmap_vmin, cmap_vmax
        if cmap_name is not None and None not in cmap_range:
            # noinspection PyTypeChecker
            return cmap_name, cmap_range

        ds = self.get_dataset(ds_id, expected_var_names=[var_name])
        var = ds[var_name]
        valid_range = get_var_valid_range(var)
        return get_var_cmap_params(var, cmap_name, cmap_range, valid_range)

    def _get_cm_styles(self) -> Tuple[Dict[str, Any], ColormapRegistry]:
        custom_colormaps = {}
        cm_styles = {}
        for style in self.config.get("Styles", []):
            style_id = style["Identifier"]
            color_mappings = dict()
            for var_name, color_mapping in style["ColorMappings"].items():
                if "ColorFile" not in color_mapping:
                    color_mappings[var_name] = dict(color_mapping)
                    continue
                custom_cmap_path = self.get_config_path(
                    color_mapping,
                    "ColorMappings",
                    path_entry_name="ColorFile"
                )
                custom_colormap = custom_colormaps.get(custom_cmap_path)
                if custom_colormap is None:
                    custom_colormap = load_custom_colormap(
                        custom_cmap_path
                    )
                    if custom_colormap is not None:
                        custom_colormaps[custom_cmap_path] = custom_colormap
                if custom_colormap is not None:
                    color_mappings[var_name] = {
                        "ColorBar": custom_colormap.cm_name,
                        "ValueRange": (custom_colormap.norm.vmin,
                                       custom_colormap.norm.vmax)
                    }
            cm_styles[style_id] = color_mappings

        return cm_styles, ColormapRegistry(*custom_colormaps.values())

    def get_color_mappings(self, ds_id: str) \
            -> Optional[Dict[str, Dict[str, Any]]]:
        dataset_config = self.get_dataset_config(ds_id)
        style_id = dataset_config.get('Style', 'default')
        return self._cm_styles.get(style_id, {})

    def _get_dataset_entry(self, ds_id: str) \
            -> Tuple[MultiLevelDataset, ServerConfig]:
        if ds_id not in self._dataset_cache:
            with self.rlock:
                self._set_dataset_entry(self._create_dataset_entry(ds_id))
        return self._dataset_cache[ds_id]

    def _set_dataset_entry(
            self,
            dataset_entry: Tuple[MultiLevelDataset, DatasetConfig]
    ):
        ml_dataset, dataset_config = dataset_entry
        self._dataset_cache[ml_dataset.ds_id] = ml_dataset, dataset_config

    def _create_dataset_entry(self, ds_id: str) \
            -> Tuple[MultiLevelDataset, DatasetConfig]:
        dataset_config = self.get_dataset_config(ds_id)
        ml_dataset = self._open_ml_dataset(dataset_config)
        return ml_dataset, dataset_config

    def _open_ml_dataset(self, dataset_config: DatasetConfig) \
            -> MultiLevelDataset:
        ds_id: str = dataset_config.get('Identifier')
        store_instance_id = dataset_config.get('StoreInstanceId')
        if store_instance_id:
            data_store_pool = self.get_data_store_pool()
            data_store = data_store_pool.get_store(store_instance_id)
            data_id = dataset_config.get('Path')
            open_params = dataset_config.get('StoreOpenParams') or {}
            # Inject chunk_cache_capacity into open parameters
            chunk_cache_capacity = self.get_dataset_chunk_cache_capacity(
                dataset_config
            )
            if chunk_cache_capacity \
                    and (data_id.endswith('.zarr')
                         or data_id.endswith('.levels')) \
                    and 'cache_size' not in open_params:
                open_params['cache_size'] = chunk_cache_capacity
            with self.measure_time(tag=f"Opened dataset {ds_id!r}"
                                       f" from data store"
                                       f" {store_instance_id!r}"):
                dataset = data_store.open_data(data_id, **open_params)
            if isinstance(dataset, MultiLevelDataset):
                ml_dataset: MultiLevelDataset = dataset
            else:
                ml_dataset = BaseMultiLevelDataset(dataset)
            ml_dataset.ds_id = ds_id
        else:
            fs_type = dataset_config.get('FileSystem')
            if fs_type != 'memory':
                raise ApiError.InvalidServerConfig(
                    f"Invalid FileSystem {fs_type!r}"
                    f" in dataset configuration"
                    f" {ds_id!r}"
                )
            with self.measure_time(tag=f"Opened dataset {ds_id!r}"
                                       f" from {fs_type!r}"):
                ml_dataset = _open_ml_dataset_from_python_code(
                    self, dataset_config
                )
        augmentation = dataset_config.get('Augmentation')
        if augmentation:
            script_path = self.get_config_path(
                augmentation,
                f"'Augmentation' of dataset configuration {ds_id}"
            )
            input_parameters = augmentation.get('InputParameters', {})
            input_parameters = self.eval_config_value(input_parameters)
            callable_name = augmentation.get('Class')
            is_factory = callable_name is not None
            if not is_factory:
                callable_name = augmentation.get('Function', COMPUTE_VARIABLES)
            ml_dataset = augment_ml_dataset(
                ml_dataset,
                script_path,
                callable_name,
                self.get_ml_dataset,
                self.set_ml_dataset,
                input_parameters=input_parameters,
                is_factory=is_factory,
                exception_type=ApiError.InvalidServerConfig
            )
        return ml_dataset

    def get_legend_label(self, ds_id: str, var_name: str):
        dataset = self.get_dataset(ds_id)
        if var_name in dataset:
            ds = self.get_dataset(ds_id)
            units = ds[var_name].units
            return units
        raise ApiError.NotFound(
            f'Variable "{var_name}" not found in dataset "{ds_id}"'
        )

    def get_dataset_place_groups(self,
                                 ds_id: str,
                                 base_url: str,
                                 load_features=False) -> List[Dict]:
        dataset_config = self.get_dataset_config(ds_id)

        place_group_id_prefix = f"DS-{ds_id}-"

        def predicate(pg_id: str, _: Any):
            return pg_id.startswith(place_group_id_prefix)

        place_groups = self.places_ctx.get_cached_place_groups(
            predicate=predicate
        )
        if place_groups:
            # for a given dataset, all place groups cached,
            # so the returned list is complete
            return place_groups

        place_groups = self.places_ctx.load_place_groups(
            dataset_config.get("PlaceGroups", []),
            base_url,
            is_global=False,
            load_features=load_features,
            qualifiers=[ds_id]
        )

        for place_group in place_groups:
            place_group_id = place_group_id_prefix + place_group["id"]
            self.places_ctx.set_cached_place_group(place_group_id,
                                                   place_group)

        return place_groups

    def get_dataset_place_group(self,
                                ds_id: str,
                                place_group_id: str,
                                base_url: str,
                                load_features=False) -> Dict:
        place_groups = self.get_dataset_place_groups(ds_id, base_url,
                                                     load_features=False)
        for place_group in place_groups:
            if place_group_id == place_group['id']:
                if load_features:
                    self.places_ctx.load_place_group_features(place_group)
                return place_group
        raise ApiError.NotFound(
            f'Place group "{place_group_id}" not found')

    def get_dataset_and_coord_variable(self, ds_name: str, dim_name: str):
        ds = self.get_dataset(ds_name)
        if dim_name not in ds.coords:
            raise ApiError.NotFound(
                f'Dimension {dim_name!r} has no coordinates'
                f' in dataset {ds_name!r}'
            )
        return ds, ds.coords[dim_name]

    @classmethod
    def get_var_indexers(cls,
                         ds_name: str,
                         var_name: str,
                         var: xr.DataArray,
                         dim_names: List[str],
                         query_params: Dict[str, str]) -> Dict[str, Any]:
        var_indexers = dict()
        for dim_name in dim_names:
            if dim_name not in var.coords:
                raise ApiError.BadRequest(
                    f'Dimension {dim_name!r} of variable {var_name!r}'
                    f' in dataset {ds_name!r} has no coordinates'
                )
            coord_var = var.coords[dim_name]
            dim_value_str = query_params.get(dim_name)
            try:
                if dim_value_str is None:
                    var_indexers[dim_name] = coord_var.values[0]
                elif dim_value_str == 'current':
                    var_indexers[dim_name] = coord_var.values[-1]
                elif np.issubdtype(coord_var.dtype, np.floating):
                    var_indexers[dim_name] = float(dim_value_str)
                elif np.issubdtype(coord_var.dtype, np.integer):
                    var_indexers[dim_name] = int(dim_value_str)
                elif np.issubdtype(coord_var.dtype, np.datetime64):
                    if '/' in dim_value_str:
                        date_str_1, date_str_2 = dim_value_str.split(
                            '/', maxsplit=1
                        )
                        var_indexer_1 = pd.to_datetime(date_str_1)
                        var_indexer_2 = pd.to_datetime(date_str_2)
                        var_indexers[dim_name] = var_indexer_1 + (
                                var_indexer_2 - var_indexer_1) / 2
                    else:
                        date_str = dim_value_str
                        var_indexers[dim_name] = pd.to_datetime(date_str)
                else:
                    raise ValueError(
                        f'unable to convert value'
                        f' {dim_value_str!r} to {coord_var.dtype!r}'
                    )
            except ValueError as e:
                raise ApiError.BadRequest(
                    f'{dim_value_str!r} is not a valid value'
                    f' for dimension {dim_name!r} '
                    f'of variable {var_name!r} of dataset {ds_name!r}'
                ) from e
        return var_indexers

    def get_dataset_chunk_cache_capacity(
            self,
            dataset_config: DatasetConfig
    ) -> Optional[int]:
        cache_size = self.get_chunk_cache_capacity(
            dataset_config, 'ChunkCacheSize'
        )
        if cache_size is None:
            cache_size = self.get_chunk_cache_capacity(
                self.config, 'DatasetChunkCacheSize'
            )
        return cache_size

    @classmethod
    def get_chunk_cache_capacity(
            cls,
            config: Mapping[str, Any],
            cache_size_key: str
    ) -> Optional[int]:
        cache_size = config.get(cache_size_key, None)
        if not cache_size:
            return None
        elif isinstance(cache_size, str):
            try:
                cache_size = parse_mem_size(cache_size)
            except ValueError:
                raise ApiError.InvalidServerConfig(
                    f'Invalid {cache_size_key}'
                )
        elif not isinstance(cache_size, int) or cache_size < 0:
            raise ApiError.InvalidServerConfig(f'Invalid {cache_size_key}')
        return cache_size


def _open_ml_dataset_from_python_code(
        ctx: DatasetsContext,
        dataset_config: DatasetConfig
) -> MultiLevelDataset:
    ds_id = dataset_config.get('Identifier')
    path = ctx.get_config_path(dataset_config,
                               f"dataset configuration {ds_id}")
    callable_name = dataset_config.get('Class')
    is_factory = callable_name is not None
    if not is_factory:
        callable_name = dataset_config.get('Function', COMPUTE_DATASET)
    input_dataset_ids = dataset_config.get('InputDatasets', [])
    input_parameters = dataset_config.get('InputParameters', {})
    input_parameters = ctx.eval_config_value(input_parameters)
    chunk_cache_capacity = ctx.get_dataset_chunk_cache_capacity(
        dataset_config
    )
    if chunk_cache_capacity:
        warnings.warn(
            'chunk cache size is not effective for'
            ' datasets computed from scripts'
        )
    for input_dataset_id in input_dataset_ids:
        if not ctx.get_dataset_config(input_dataset_id):
            raise ApiError.InvalidServerConfig(
                f"Invalid dataset configuration {ds_id!r}:"
                f" Input dataset {input_dataset_id!r} of"
                f" callable {callable_name!r}"
                f" must reference another dataset"
            )
    return open_ml_dataset_from_python_code(
        path,
        callable_name=callable_name,
        input_ml_dataset_ids=input_dataset_ids,
        input_ml_dataset_getter=ctx.get_ml_dataset,
        input_parameters=input_parameters,
        is_factory=is_factory,
        ds_id=ds_id,
        exception_type=ApiError.InvalidServerConfig
    )


def _is_not_empty(prefix):
    return prefix != '' and prefix != '/' and prefix != '\\'


def _get_roots(paths):
    list.sort(paths)
    prefixes = _get_common_prefixes(paths)
    if len(prefixes) < 1:
        roots = ['']
    else:
        # perform further step to merge prefixes with same start
        prefixes = list(set(prefixes))
        prefixes.sort()
        roots = []
        root_candidate = prefixes[0]
        for root in prefixes[1:]:
            common_root = os.path.commonprefix([root_candidate, root])
            if _is_not_empty(common_root):
                root_candidate = common_root
            else:
                roots.append(root_candidate)
                root_candidate = root
        roots.append(root_candidate)
    if roots[0] == '':
        roots.append(roots.pop(0))
    return roots


def _get_common_prefixes(p):
    # Recursively examine a list of paths for common prefixes:
    # If no common prefix is found, split the list in half and
    # examine each half separately
    prefix = os.path.commonprefix(p)
    # ensure prefix does not end with full or partial directory
    # or file name
    prefix = prefix[:max(_lastindex(prefix, '/'), _lastindex(prefix, '\\'), 0)]
    if _is_not_empty(prefix) or len(p) == 1:
        return [prefix]
    else:
        return _get_common_prefixes(p[:int(len(p) / 2)]) + \
               _get_common_prefixes(p[int(len(p) / 2):])


def _lastindex(prefix, symbol):
    try:
        return prefix.rindex(symbol)
    except ValueError:
        return -1


_MULTI_LEVEL_DATASET_OPENERS = {
    "memory": _open_ml_dataset_from_python_code,
}



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click

DEFAULT_OUTPUT_PATH = '{input}-optimized.zarr'


# noinspection PyShadowingBuiltins
@click.command(name='optimize')
@click.argument('CUBE')
@click.option('--output', '-o', metavar='OUTPUT',
              help=f'Output path. The placeholder "{input}" will be replaced by the input\'s filename '
                   f'without extension (such as ".zarr"). Defaults to "{DEFAULT_OUTPUT_PATH}".',
              default=DEFAULT_OUTPUT_PATH)
@click.option('--in-place', '-I', 'in_place',
              help="Optimize cube in place. Ignores output path.",
              is_flag=True)
@click.option('--coords', '-C', 'unchunk_coords',
              help="Also optimize coordinate variables by converting any chunked arrays "
                   "into single, non-chunked, contiguous arrays.",
              is_flag=True)
def optimize(cube,
             output,
             in_place,
             unchunk_coords):
    """
    Optimize xcube dataset for faster access.

    Reduces the number of metadata and coordinate data files in xcube dataset given by CUBE.
    Consolidated cubes open much faster especially from remote locations, e.g. in object storage,
    because obviously much less HTTP requests are required to fetch initial cube meta
    information. That is, it merges all metadata files into a single top-level JSON file ".zmetadata".
    Optionally, it removes any chunking of coordinate variables
    so they comprise a single binary data file instead of one file per data chunk.
    The primary usage of this command is to optimize data cubes for cloud object storage.
    The command currently works only for data cubes using ZARR format.
    """
    from xcube.core.optimize import optimize_dataset
    optimize_dataset(cube,
                     output_path=output,
                     in_place=in_place,
                     unchunk_coords=unchunk_coords)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os.path
import uuid
from typing import Sequence, Any, Dict, Callable, Mapping, Optional, Tuple

import xarray as xr

from xcube.core.byoa import CodeConfig
from xcube.core.byoa import FileSet
from xcube.core.gridmapping import GridMapping
from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.perf import measure_time
from .abc import MultiLevelDataset
from .lazy import LazyMultiLevelDataset

MultiLevelDatasetGetter = Callable[[str], MultiLevelDataset]
MultiLevelDatasetSetter = Callable[[MultiLevelDataset], None]


class ComputedMultiLevelDataset(LazyMultiLevelDataset):
    """A multi-level dataset whose level datasets are computed
    by a user function.

    The script can import other Python modules located in the same
    directory as *script_path*.
    """

    def __init__(self,
                 script_path: str,
                 callable_name: str,
                 input_ml_dataset_ids: Sequence[str],
                 input_ml_dataset_getter: MultiLevelDatasetGetter,
                 input_parameters: Optional[Mapping[str, Any]] = None,
                 ds_id: str = '',
                 exception_type: type = ValueError):

        callable_ref, callable_obj = self.get_callable(
            script_path,
            callable_name,
            input_ml_dataset_ids,
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            ds_id=ds_id,
            exception_type=exception_type
        )

        super().__init__(ds_id=ds_id, parameters=input_parameters)
        self._callable_ref = callable_ref
        self._callable_obj = callable_obj
        self._input_ml_dataset_ids = input_ml_dataset_ids
        self._input_ml_dataset_getter = input_ml_dataset_getter
        self._exception_type = exception_type

    @classmethod
    def get_callable(
            cls,
            script_path: str,
            callable_name: str,
            input_ml_dataset_ids: Sequence[str],
            input_ml_dataset_getter: MultiLevelDatasetGetter,
            input_parameters: Optional[Mapping[str, Any]] = None,
            ds_id: str = '',
            exception_type: type = ValueError
    ) -> Tuple[str, Callable]:

        assert_instance(script_path, str, name='script_path')
        assert_given(script_path, name='script_path')
        assert_true(callable(input_ml_dataset_getter),
                    message=f'input_ml_dataset_getter must be a callable')
        assert_given(input_ml_dataset_getter, name='input_ml_dataset_getter')
        assert_instance(ds_id, str, name='ds_id')
        assert_given(ds_id, name='ds_id')

        module_name = None
        basename = os.path.basename(script_path)
        basename, ext = os.path.splitext(basename)
        if ext == '.py':
            script_path = os.path.dirname(script_path)
            module_name = basename

        if ':' in callable_name:
            callable_ref = callable_name
        else:
            if not module_name:
                raise exception_type(
                    f"Invalid in-memory dataset descriptor {ds_id!r}:"
                    f' Missing module name in {callable_name!r}'
                )
            callable_ref = f'{module_name}:{callable_name}'

        if not input_ml_dataset_ids:
            raise exception_type(
                f"Invalid in-memory dataset descriptor {ds_id!r}:"
                f" Input dataset(s) missing for callable {callable_name!r}"
            )

        for input_param_name in (input_parameters or {}).keys():
            if not input_param_name or not input_param_name.isidentifier():
                raise exception_type(
                    f"Invalid in-memory dataset descriptor {ds_id!r}:"
                    f" Input parameter {input_param_name!r}"
                    f" for callable {callable_name!r}"
                    f" is not a valid Python identifier"
                )

        try:
            callable_obj = CodeConfig.from_file_set(
                FileSet(path=script_path),
                callable_ref=callable_ref,
                install_required=False
            ).get_callable()
        except (TypeError, ValueError, ImportError) as e:
            raise exception_type(
                f"Invalid dataset descriptor {ds_id!r}: {e}"
            ) from e

        return callable_ref, callable_obj

    @property
    def num_inputs(self) -> int:
        return len(self._input_ml_dataset_ids)

    def get_input_dataset(self, index: int) -> MultiLevelDataset:
        return self._input_ml_dataset_getter(
            self._input_ml_dataset_ids[index]
        )

    def _get_num_levels_lazily(self) -> int:
        return self.get_input_dataset(0).num_levels

    def _get_grid_mapping_lazily(self) -> GridMapping:
        return self.get_input_dataset(0).grid_mapping

    def _get_dataset_lazily(self,
                            index: int,
                            parameters: Dict[str, Any]) -> xr.Dataset:
        input_datasets = [
            self._input_ml_dataset_getter(ds_id).get_dataset(index)
            for ds_id in self._input_ml_dataset_ids]
        try:
            with measure_time(tag=f"Computed in-memory dataset"
                                  f" {self.ds_id!r} at level {index}"):
                computed_value = self._callable_obj(*input_datasets,
                                                    **parameters)
        except Exception as e:
            raise self._exception_type(
                f"Failed to compute in-memory dataset {self.ds_id!r}"
                f" at level {index} "
                f"from function {self._callable_ref!r}(): {e}"
            ) from e
        if not isinstance(computed_value, xr.Dataset):
            raise self._exception_type(
                f"Failed to compute in-memory dataset {self.ds_id!r}"
                f" at level {index} "
                f"from function {self._callable_ref!r}(): "
                f"expected an xarray.Dataset but got {type(computed_value)}"
            )
        return computed_value


def augment_ml_dataset(
        ml_dataset: MultiLevelDataset,
        script_path: str,
        callable_name: str,
        input_ml_dataset_getter: MultiLevelDatasetGetter,
        input_ml_dataset_setter: MultiLevelDatasetSetter,
        input_parameters: Optional[Mapping[str, Any]] = None,
        is_factory: bool = False,
        exception_type: type = ValueError
):
    from .identity import IdentityMultiLevelDataset
    from .combined import CombinedMultiLevelDataset
    with measure_time(tag=f"Added augmentation from {script_path}"):
        orig_id = ml_dataset.ds_id
        aug_id = uuid.uuid4()
        aug_inp_id = f'aug-input-{aug_id}'
        aug_inp_ds = IdentityMultiLevelDataset(ml_dataset, ds_id=aug_inp_id)
        input_ml_dataset_setter(aug_inp_ds)
        aug_ds = _open_ml_dataset_from_python_code(
            script_path,
            callable_name,
            [aug_inp_id],
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            is_factory=is_factory,
            ds_id=f'aug-{aug_id}',
            exception_type=exception_type
        )
        return CombinedMultiLevelDataset([ml_dataset, aug_ds], ds_id=orig_id)


def open_ml_dataset_from_python_code(
        script_path: str,
        callable_name: str,
        input_ml_dataset_ids: Sequence[str],
        input_ml_dataset_getter: MultiLevelDatasetGetter,
        input_parameters: Optional[Mapping[str, Any]] = None,
        is_factory: bool = False,
        ds_id: str = '',
        exception_type: type = ValueError
) -> MultiLevelDataset:
    with measure_time(tag=f"Opened memory dataset {script_path}"):
        return _open_ml_dataset_from_python_code(
            script_path,
            callable_name,
            input_ml_dataset_ids,
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            is_factory=is_factory,
            ds_id=ds_id,
            exception_type=exception_type
        )


def _open_ml_dataset_from_python_code(
        script_path: str,
        callable_name: str,
        input_ml_dataset_ids: Sequence[str],
        input_ml_dataset_getter: MultiLevelDatasetGetter,
        input_parameters: Optional[Mapping[str, Any]] = None,
        is_factory: bool = False,
        ds_id: str = '',
        exception_type: type = ValueError
) -> MultiLevelDataset:
    if is_factory:
        callable_ref, callable_obj = ComputedMultiLevelDataset.get_callable(
            script_path,
            callable_name,
            input_ml_dataset_ids,
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            ds_id=ds_id,
            exception_type=exception_type
        )
        input_datasets = [input_ml_dataset_getter(ds_id)
                          for ds_id in input_ml_dataset_ids]
        try:
            ml_dataset = callable_obj(*input_datasets,
                                      **(input_parameters or {}))
            if not isinstance(ml_dataset, MultiLevelDataset):
                raise TypeError(
                    f"{callable_ref!r} must return instance of"
                    f" xcube.core.mldataset.MultiLevelDataset,"
                    f" but was {type(ml_dataset)}"
                )
            ml_dataset.ds_id = ds_id
            return ml_dataset
        except BaseException as e:
            raise exception_type(
                f"Invalid in-memory dataset descriptor {ds_id!r}: {e}"
            ) from e
    else:
        return ComputedMultiLevelDataset(
            script_path,
            callable_name,
            input_ml_dataset_ids,
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            ds_id=ds_id,
            exception_type=exception_type
        )



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import os.path
from typing import List, Sequence

import numpy as np
import xarray as xr
import zarr


def unchunk_dataset(dataset_path: str, var_names: Sequence[str] = None, coords_only: bool = False):
    """
    Unchunk dataset variables in-place.

    :param dataset_path: Path to ZARR dataset directory.
    :param var_names: Optional list of variable names.
    :param coords_only: Un-chunk coordinate variables only.
    """

    is_zarr = os.path.isfile(os.path.join(dataset_path, '.zgroup'))
    if not is_zarr:
        raise ValueError(f'{dataset_path!r} is not a valid Zarr directory')

    with xr.open_zarr(dataset_path) as dataset:
        if var_names is None:
            if coords_only:
                var_names = list(dataset.coords)
            else:
                var_names = list(dataset.variables)
        else:
            for var_name in var_names:
                if coords_only:
                    if var_name not in dataset.coords:
                        raise ValueError(f'variable {var_name!r} is not a coordinate variable in {dataset_path!r}')
                else:
                    if var_name not in dataset.variables:
                        raise ValueError(f'variable {var_name!r} is not a variable in {dataset_path!r}')

    _unchunk_vars(dataset_path, var_names)


def _unchunk_vars(dataset_path: str, var_names: List[str]):
    for var_name in var_names:
        var_path = os.path.join(dataset_path, var_name)

        # Optimization: if "shape" and "chunks" are equal in ${var}/.zarray, we are done
        var_array_info_path = os.path.join(var_path, '.zarray')
        with open(var_array_info_path, 'r') as fp:
            var_array_info = json.load(fp)
            if var_array_info.get('shape') == var_array_info.get('chunks'):
                continue

        # Open array and remove chunks from the data
        var_array = zarr.convenience.open_array(var_path, 'r+')
        if var_array.shape != var_array.chunks:
            # TODO (forman): Fully loading data is inefficient and dangerous for large arrays.
            #                Instead save unchunked to temp and replace existing chunked array dir with temp.
            # Fully load data and attrs so we no longer depend on files
            data = np.array(var_array)
            attributes = var_array.attrs.asdict()
            # Save array data
            zarr.convenience.save_array(var_path, data, chunks=False, fill_value=var_array.fill_value)
            # zarr.convenience.save_array() does not seem save user attributes (file ".zattrs" not written),
            # therefore we must modify attrs explicitly:
            var_array = zarr.convenience.open_array(var_path, 'r+')
            var_array.attrs.update(attributes)

    zarr.consolidate_metadata(dataset_path)



import collections.abc
import inspect
import typing
from typing import Callable, Dict, Any
from mashumaro.jsonschema import build_json_schema

import xarray as xr

from xcube.core.mldataset import MultiLevelDataset

PyType = type(type(int))

# Functions registered in the operation registry will
# receive a new attribute with this name.
_ATTR_NAME_OP_INFO = '_op_info'

_PRIMITIVE_PY_TO_JSON_TYPES = {
    type(None): "null",
    bool: "boolean",
    int: "integer",
    float: "number",
    str: "string",
}

_PRIMITIVE_JSON_TO_PY_TYPES = {
    v: k for k, v in _PRIMITIVE_PY_TO_JSON_TYPES.items()
}


class OpInfo:
    """Information about a compute operation"""

    def __init__(self,
                 params_schema: Dict[str, Any],
                 param_py_types: Dict[str, PyType]):
        """Create information about a compute operation

        :param params_schema: map of parameter names to their JSON Schema
                              definitions
        :param param_py_types: map of parameter names to their Python types
        """
        self.params_schema = params_schema
        self.param_py_types = param_py_types

    def make_op(self, function: Callable) -> Callable:
        """Add this information instance to a function

        The supplied function is modified in-place (with the addition of an
        attribute referencing this information instance) and returned.

        :param function: the function to associate with this information
        :return: the same function (with an attribute added)
        """
        setattr(function, _ATTR_NAME_OP_INFO, self)
        return function

    @classmethod
    def get_op_info(cls, op: Callable) -> "OpInfo":
        """Get the information object for a specified function

        :param op: the function for which information is requested
        :return: the function’s associated information object, if present
        :raises ValueError: if there is no associated information object
                            (i.e. the function is not an operation)
        """
        op_info = getattr(op, _ATTR_NAME_OP_INFO, None)
        if not isinstance(op_info, OpInfo):
            raise ValueError(f"function {op.__name__}() is not"
                             f" registered as an operation")
        return op_info

    @classmethod
    def new_op_info(cls, op: Callable) -> "OpInfo":
        """Create a new operation information object for a function

        The returned information object contains a parameter type dictionary
        and JSON schema created by analysis of the supplied function.

        :param op: a function
        :return: operation information for the supplied function
        """
        members = dict(inspect.getmembers(op))
        annotations = members.get("__annotations__")
        code = members.get("__code__")
        param_py_types = {}
        params_schema = {}
        if code:
            args = inspect.getargs(code)
            required_param_names = set(args.args or []) \
                .union(set(args.varargs or []))
            optional_param_names = set(args.varkw or [])
            all_param_names = required_param_names.union(optional_param_names)
            if all_param_names:
                properties = {}
                for param_name in all_param_names:
                    py_type = annotations.get(param_name)
                    if py_type is not None:
                        param_py_types[param_name] = py_type
                    if inspect.isclass(py_type):
                        if issubclass(py_type,
                                      (xr.Dataset, MultiLevelDataset)):
                            # extract function from get_effective_parameters
                            param_schema = {
                                "type": "string",
                                "title": "Dataset identifier"
                            }
                        else:
                            raise ValueError(
                                f'Illegal operation parameter class {py_type}.'
                                f' Classes must be subclasses of Dataset or '
                                f'MultiLevelDataset.'
                            )
                    elif py_type is not None:
                        if not OpInfo._is_valid_parameter_type(py_type):
                            raise ValueError(
                                f'Illegal operation parameter type {py_type}.'
                            )
                        param_schema = build_json_schema(py_type).to_dict()
                    else:
                        param_schema = {}
                    properties[param_name] = param_schema
                params_schema = {
                    "type": "object",
                    "properties": properties,
                    "required": list(required_param_names),
                    "additionalProperties": False
                }
            else:
                params_schema.update({
                    "type": ["null", "object"],
                    "additionalProperties": False
                })

        return OpInfo(params_schema, param_py_types)

    @staticmethod
    def _is_valid_parameter_type(py_type) -> bool:
        """Raise an exception if the supplied type is not valid for operations

        "Valid" means that it is composed only of sequences, mappings,
        primitives, and combinations thereof.
        """
        pass
        origin = typing.get_origin(py_type)
        args = typing.get_args(py_type)
        if py_type in [int, float, str]:
            return True
        elif origin in [dict, collections.abc.Mapping]:
            return args[0] is str and OpInfo._is_valid_parameter_type(args[1])
        elif origin in [list, tuple, collections.abc.Sequence]:
            return all(map(OpInfo._is_valid_parameter_type, args))

    @property
    def effective_param_py_types(self) -> Dict[str, PyType]:
        """Return effective Python types for the operation’s parameters

        For any parameter which has a Python type annotation, that type will
        be set as the value in the returned mapping. If a parameter has no
        Python type annotation but a primitive type is defined in the
        operation’s JSON schema, the corresponding Python type will be used
        as the value in the returned type dictionary. Otherwise `None` will
        be used.

        :return: a dictionary mapping operation parameter names to their
                 effective Python types
        """
        py_types = self.param_py_types.copy()
        for param_name, param_schema in self.param_schemas.items():
            py_type = py_types.get(param_name)
            if py_type is None:
                json_type = param_schema.get("type")
                # TODO Document requirement that operations must have type
                #  annotations
                # TODO check whether matumosho can do the inverse transform
                #  for us
                py_type = _PRIMITIVE_JSON_TO_PY_TYPES.get(json_type)
                if py_type is not None:
                    py_types[param_name] = py_type
        return py_types

    @property
    def param_schemas(self) -> Dict[str, Any]:
        """
        :return: a mapping of parameter names to their JSON schemas
        """
        return self.params_schema.get("properties", {})

    def set_param_schemas(self, schemas: Dict[str, Any]):
        """Set JSON schemas for operation parameters

        :param schemas: a mapping of parameter names to their JSON schemas
        """
        self.params_schema["properties"] = schemas

    def update_params_schema(self, schema: Dict[str, Any]):
        """Update the JSON schema for the whole parameters dictionary

        Update the existing parameter dictionary schema with the supplied
        dictionary. The supplied dictionary should be a valid JSON Schema,
        with any schemas for the individual parameters in a sub-dictionary
        under the `properties` key.

        :param schema: the schema with which to update the parameters
                       dictionary schema
        """
        self.params_schema.update(schema)

    def update_param_schema(self, param_name: str, schema: Dict[str, Any]):
        """Update the JSON Schema for a single parameter

        :param param_name: name of the parameter
        :param schema: schema with which to update the parameter’s current
                       schema
        """
        param_schemas = self.param_schemas
        param_schema = param_schemas.get(param_name, {})
        param_schema.update(schema)
        # next line needed in case default value was taken in .get above
        param_schemas[param_name] = param_schema
        self.set_param_schemas(param_schemas)

    def get_param_py_type(self, param_name: str) -> PyType:
        """Get the Python type for a parameter

        :param param_name: name of parameter
        :return: Python type of specified parameter
        """
        return self.param_py_types[param_name]

    def set_param_py_type(self, param_name: str, py_type: PyType):
        """Set the Python type for a parameter

        :param param_name: name of parameter
        :param py_type: Python type of specified parameter
        """
        self.param_py_types[param_name] = py_type



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Any, Dict, Callable

import xarray as xr

from .abc import MultiLevelDataset
from .lazy import LazyMultiLevelDataset


class MappedMultiLevelDataset(LazyMultiLevelDataset):
    def __init__(
            self,
            ml_dataset: MultiLevelDataset,
            mapper_function: Callable[[xr.Dataset], xr.Dataset],
            ds_id: str = None,
            mapper_params: Dict[str, Any] = None
    ):
        """

        """
        super().__init__(ds_id=ds_id,
                         parameters=mapper_params)
        self._ml_dataset = ml_dataset
        self._mapper_function = mapper_function

    def _get_num_levels_lazily(self) -> int:
        return self._ml_dataset.num_levels

    def _get_dataset_lazily(self, index: int,
                            mapper_params: Dict[str, Any]) -> xr.Dataset:
        return self._mapper_function(self._ml_dataset.get_dataset(index),
                                     **mapper_params)

    def close(self):
        self._ml_dataset.close()



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import Tuple

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.normalize import encode_cube
from xcube.core.store import DataStorePool
from xcube.core.store import get_data_store_instance
from xcube.core.store import new_data_writer
from xcube.util.progress import observe_dask_progress
from ..config import OutputConfig


class CubeWriter:
    def __init__(self,
                 output_config: OutputConfig,
                 store_pool: DataStorePool = None):
        self._output_config = output_config
        self._store_pool = store_pool

    def write_cube(self,
                   cube: xr.Dataset,
                   gm: GridMapping) -> Tuple[str, xr.Dataset]:
        output_config = self._output_config
        dataset = encode_cube(cube, grid_mapping=gm)
        with observe_dask_progress('writing cube', 100):
            write_params = output_config.write_params or {}
            store_params = output_config.store_params or {}
            if output_config.store_id:
                store_instance = get_data_store_instance(
                    output_config.store_id,
                    store_params=store_params,
                    store_pool=self._store_pool
                )
                writer = store_instance.store
                write_params.update(
                    writer_id=output_config.writer_id,
                    **write_params
                )
            else:
                writer = new_data_writer(output_config.writer_id)
                write_params.update(**store_params, **write_params)
            if not dataset.attrs.get('title'):
                # Set fallback title, so we can distinguish
                # datasets from stores in xcube-viewer
                dataset = dataset.assign_attrs(title=output_config.data_id)
            data_id = writer.write_data(
                dataset,
                data_id=output_config.data_id,
                replace=output_config.replace or False,
                **write_params
            )
        return data_id, dataset
